#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_campus_gateway_clusters
short_description: Resource module for networks _campus _gateway _clusters
description:
  - Manage operations create and update of the resource networks _campus _gateway
    _clusters.
  - Create a cluster and add campus gateways to it.
  - Update a cluster and add/remove campus gateways to/from it.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  clusterId:
    description: ClusterId path parameter. Cluster ID.
    type: str
  devices:
    description: Devices to be added to the cluster.
    elements: dict
    suboptions:
      serial:
        description: Serial of the device.
        type: str
      tunnels:
        description: Tunnel settings of the device when tunnel interface of cluster
          is specified.
        elements: dict
        suboptions:
          addresses:
            description: Tunnel IP addresses of the device.
            elements: dict
            suboptions:
              address:
                description: IP address of the interface.
                type: str
              protocol:
                description: Protocol of the interface.
                type: str
            type: list
          interface:
            description: Tunnel interface name.
            type: str
        type: list
      uplinks:
        description: Uplink settings of the device when uplink assignment mode of
          cluster is static.
        elements: dict
        suboptions:
          addresses:
            description: Uplink IP addresses of the device.
            elements: dict
            suboptions:
              address:
                description: IP address of the interface.
                type: str
              protocol:
                description: Protocol of the interface.
                type: str
            type: list
          interface:
            description: Uplink interface name.
            type: str
        type: list
    type: list
  name:
    description: Name of the new cluster.
    type: str
  nameservers:
    description: Nameservers of the cluster.
    suboptions:
      addresses:
        description: Addresses of the nameservers.
        elements: str
        type: list
    type: dict
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  notes:
    description: Notes about cluster with max size of 511 characters allowed.
    type: str
  portChannels:
    description: Port channel settings of the cluster.
    elements: dict
    suboptions:
      allowedVlans:
        description: Allowed VLANs of the port channel.
        type: str
      name:
        description: Name of the port channel.
        type: str
      vlan:
        description: VLAN ID of the port channel.
        type: int
    type: list
  tunnels:
    description: Tunnel interface settings of the cluster Reuse uplink or specify
      tunnel interface.
    elements: dict
    suboptions:
      addresses:
        description: Addresses of the interface.
        elements: dict
        suboptions:
          gateway:
            description: Gateway of the interface.
            type: str
          protocol:
            description: Protocol of the interface.
            type: str
          subnetMask:
            description: Subnet mask of the interface.
            type: str
        type: list
      interface:
        description: Interface identifier, should be set to tun1. Specify this interface
          only if uplink isn't reused as tunnel.
        type: str
      uplink:
        description: Uplink interface details if uplink is reused as tunnel.
        suboptions:
          interface:
            description: Uplink interface identifier.
            type: str
        type: dict
      vlan:
        description: VLAN ID of the interface.
        type: int
    type: list
  uplinks:
    description: Uplink interface settings of the cluster.
    elements: dict
    suboptions:
      addresses:
        description: Addresses of the interface.
        elements: dict
        suboptions:
          assignmentMode:
            description: Assignment mode of the interface.
            type: str
          gateway:
            description: Gateway of the interface when assignment mode is static.
            type: str
          protocol:
            description: Protocol of the interface when assignment mode is static.
            type: str
          subnetMask:
            description: Subnet mask of the interface when assignment mode is static.
            type: str
        type: list
      interface:
        description: Interface identifier, should be set to man1.
        type: str
      vlan:
        description: VLAN ID of the interface.
        type: int
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for campusGateway createNetworkCampusGatewayCluster
    description: Complete reference of the createNetworkCampusGatewayCluster API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-network-campus-gateway-cluster
  - name: Cisco Meraki documentation for campusGateway updateNetworkCampusGatewayCluster
    description: Complete reference of the updateNetworkCampusGatewayCluster API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-campus-gateway-cluster
notes:
  - SDK Method used are
    campus_gateway.CampusGateway.create_network_campus_gateway_cluster,
    campus_gateway.CampusGateway.update_network_campus_gateway_cluster,
  - Paths used are
    post /networks/{networkId}/campusGateway/clusters,
    put /networks/{networkId}/campusGateway/clusters/{clusterId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.networks_campus_gateway_clusters:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    devices:
      - serial: Q234-ABCD-0001
        tunnels:
          - addresses:
              - address: 6.2.6.7
                protocol: ipv4
            interface: tun1
        uplinks:
          - addresses:
              - address: 5.1.2.3
                protocol: ipv4
            interface: man1
    name: North Campus
    nameservers:
      addresses:
        - 8.8.8.8
        - 8.8.4.4
    networkId: string
    notes: This cluster is for New York Office
    portChannels:
      - allowedVlans: 10-20
        name: Port-channel1
        vlan: 25
    tunnels:
      - addresses:
          - gateway: 2.3.5.6
            protocol: ipv4
            subnetMask: 255.255.255.0
        interface: tun1
        uplink:
          interface: man1
        vlan: 6
    uplinks:
      - addresses:
          - assignmentMode: static
            gateway: 1.2.3.5
            protocol: ipv4
            subnetMask: 255.255.255.0
        interface: man1
        vlan: 5
- name: Update by id
  cisco.meraki.networks_campus_gateway_clusters:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    clusterId: string
    devices:
      - serial: Q234-ABCD-0001
        tunnels:
          - addresses:
              - address: 6.2.6.7
                protocol: ipv4
            interface: tun1
        uplinks:
          - addresses:
              - address: 5.1.2.3
                protocol: ipv4
            interface: man1
    name: North Campus
    nameservers:
      addresses:
        - 8.8.8.8
        - 8.8.4.4
    networkId: string
    notes: This cluster is for New York Office
    portChannels:
      - allowedVlans: 10-20
        name: Port-channel1
        vlan: 25
    tunnels:
      - addresses:
          - gateway: 2.3.5.6
            protocol: ipv4
            subnetMask: 255.255.255.0
        interface: tun1
        uplink:
          interface: man1
        vlan: 6
    uplinks:
      - addresses:
          - assignmentMode: static
            gateway: 1.2.3.5
            protocol: ipv4
            subnetMask: 255.255.255.0
        interface: man1
        vlan: 5
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "clusterId": "string",
      "name": "string",
      "uplinks": [
        {
          "interface": "string",
          "vlan": 0,
          "addresses": [
            {
              "assignmentMode": "string",
              "protocol": "string",
              "gateway": "string",
              "subnetMask": "string"
            }
          ]
        }
      ],
      "tunnels": [
        {
          "uplink": {
            "interface": "string"
          },
          "interface": "string",
          "vlan": 0,
          "addresses": [
            {
              "protocol": "string",
              "gateway": "string",
              "subnetMask": "string"
            }
          ]
        }
      ],
      "nameservers": {
        "addresses": [
          "string"
        ]
      },
      "portChannels": [
        {
          "id": "string",
          "name": "string",
          "vlan": 0,
          "allowedVlans": "string"
        }
      ],
      "devices": [
        {
          "serial": "string",
          "memberId": "string",
          "uplinks": [
            {
              "interface": "string",
              "addresses": [
                {
                  "protocol": "string",
                  "address": "string"
                }
              ]
            }
          ],
          "tunnels": [
            {
              "interface": "string",
              "addresses": [
                {
                  "protocol": "string",
                  "address": "string"
                }
              ]
            }
          ]
        }
      ],
      "notes": "string",
      "url": "string"
    }
"""



# Copyright (c) 2022 Western Digital Corporation
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import datetime
import os
import re
import tarfile
import time
import typing as t
from urllib.parse import urlparse, urlunparse

from ansible.module_utils.urls import fetch_file

from ansible_collections.community.general.plugins.module_utils._redfish_utils import RedfishUtils

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


class WdcRedfishUtils(RedfishUtils):
    """Extension to RedfishUtils to support WDC enclosures."""

    # Status codes returned by WDC FW Update Status
    UPDATE_STATUS_CODE_READY_FOR_FW_UPDATE = 0
    UPDATE_STATUS_CODE_FW_UPDATE_IN_PROGRESS = 1
    UPDATE_STATUS_CODE_FW_UPDATE_COMPLETED_WAITING_FOR_ACTIVATION = 2
    UPDATE_STATUS_CODE_FW_UPDATE_FAILED = 3

    # Status messages returned by WDC FW Update Status
    UPDATE_STATUS_MESSAGE_READY_FOR_FW_UDPATE = "Ready for FW update"
    UDPATE_STATUS_MESSAGE_FW_UPDATE_IN_PROGRESS = "FW update in progress"
    UPDATE_STATUS_MESSAGE_FW_UPDATE_COMPLETED_WAITING_FOR_ACTIVATION = "FW update completed. Waiting for activation."
    UPDATE_STATUS_MESSAGE_FW_UPDATE_FAILED = "FW update failed."

    # Dict keys for resource bodies
    # Standard keys
    ACTIONS = "Actions"
    OEM = "Oem"
    WDC = "WDC"
    TARGET = "target"

    # Keys for specific operations
    CHASSIS_LOCATE = "#Chassis.Locate"
    CHASSIS_POWER_MODE = "#Chassis.PowerMode"

    def __init__(self, creds, root_uris, timeout, module: AnsibleModule, resource_id, data_modification) -> None:
        super().__init__(
            creds=creds,
            root_uri=root_uris[0],
            timeout=timeout,
            module=module,
            resource_id=resource_id,
            data_modification=data_modification,
        )
        # Update the root URI if we cannot perform a Redfish GET to the first one
        self._set_root_uri(root_uris)

    def _set_root_uri(self, root_uris):
        """Set the root URI from a list of options.

        If the current root URI is good, just keep it.  Else cycle through our options until we find a good one.
        A URI is considered good if we can GET uri/redfish/v1.
        """
        for root_uri in root_uris:
            uri = f"{root_uri}/redfish/v1"
            response = self.get_request(uri)
            if response["ret"]:
                self.root_uri = root_uri
                break

    def _find_updateservice_resource(self):
        """Find the update service resource as well as additional WDC-specific resources."""
        response = super()._find_updateservice_resource()
        if not response["ret"]:
            return response
        return self._find_updateservice_additional_uris()

    def _is_enclosure_multi_tenant_and_fetch_gen(self):
        """Determine if the enclosure is multi-tenant.

        The serial number of a multi-tenant enclosure will end in "-A" or "-B".
        Fetching enclsoure generation.

        :return: True/False if the enclosure is multi-tenant or not and return enclosure generation;
        None if unable to determine.
        """
        response = self.get_request(f"{self.root_uri}{self.service_root}Chassis/Enclosure")
        if response["ret"] is False:
            return None
        pattern = r".*-[A,B]"
        data = response["data"]
        if "EnclVersion" not in data:
            enc_version = "G1"
        else:
            enc_version = data["EnclVersion"]
        return re.match(pattern, data["SerialNumber"]) is not None, enc_version

    def _find_updateservice_additional_uris(self):
        """Find & set WDC-specific update service URIs"""
        response = self.get_request(self.root_uri + self._update_uri())
        if response["ret"] is False:
            return response
        data = response["data"]
        if "Actions" not in data:
            return {"ret": False, "msg": "Service does not support SimpleUpdate"}
        if "#UpdateService.SimpleUpdate" not in data["Actions"]:
            return {"ret": False, "msg": "Service does not support SimpleUpdate"}
        action = data["Actions"]["#UpdateService.SimpleUpdate"]
        if "target" not in action:
            return {"ret": False, "msg": "Service does not support SimpleUpdate"}
        self.simple_update_uri = action["target"]

        # Simple update status URI is not provided via GET /redfish/v1/UpdateService
        # So we have to hard code it.
        self.simple_update_status_uri = f"{self.simple_update_uri}/Status"

        # FWActivate URI
        if "Oem" not in data["Actions"]:
            return {"ret": False, "msg": "Service does not support OEM operations"}
        if "WDC" not in data["Actions"]["Oem"]:
            return {"ret": False, "msg": "Service does not support WDC operations"}
        if "#UpdateService.FWActivate" not in data["Actions"]["Oem"]["WDC"]:
            return {"ret": False, "msg": "Service does not support FWActivate"}
        action = data["Actions"]["Oem"]["WDC"]["#UpdateService.FWActivate"]
        if "target" not in action:
            return {"ret": False, "msg": "Service does not support FWActivate"}
        self.firmware_activate_uri = action["target"]
        return {"ret": True}

    def _simple_update_status_uri(self):
        return self.simple_update_status_uri

    def _firmware_activate_uri(self):
        return self.firmware_activate_uri

    def _update_uri(self):
        return self.update_uri

    def get_simple_update_status(self):
        """Issue Redfish HTTP GET to return the simple update status"""
        result = {}
        response = self.get_request(self.root_uri + self._simple_update_status_uri())
        if response["ret"] is False:
            return response
        result["ret"] = True
        data = response["data"]
        result["entries"] = data
        return result

    def firmware_activate(self, update_opts):
        """Perform FWActivate using Redfish HTTP API."""
        creds = update_opts.get("update_creds")
        payload = {}
        if creds:
            if creds.get("username"):
                payload["Username"] = creds.get("username")
            if creds.get("password"):
                payload["Password"] = creds.get("password")

        # Make sure the service supports FWActivate
        response = self.get_request(self.root_uri + self._update_uri())
        if response["ret"] is False:
            return response
        data = response["data"]
        if "Actions" not in data:
            return {"ret": False, "msg": "Service does not support FWActivate"}

        response = self.post_request(self.root_uri + self._firmware_activate_uri(), payload)
        if response["ret"] is False:
            return response
        return {"ret": True, "changed": True, "msg": "FWActivate requested"}

    def _get_bundle_version(self, bundle_uri):
        """Get the firmware version from a bundle file, and whether or not it is multi-tenant.

        Only supports HTTP at this time.  Assumes URI exists and is a tarfile.
        Looks for a file oobm-[version].pkg, such as 'oobm-4.0.13.pkg`.  Extracts the version number
        from that filename (in the above example, the version number is "4.0.13".

        To determine if the bundle is multi-tenant or not, it looks inside the .bin file within the tarfile,
        and checks the appropriate byte in the file.

        If not tarfile, the bundle is checked for 2048th byte to determine whether it is Gen2 bundle.
        Gen2 is always single tenant at this time.

        :param str bundle_uri:  HTTP URI of the firmware bundle.
        :return: Firmware version number contained in the bundle, whether or not the bundle is multi-tenant
        and bundle generation. Either value will be None if unable to determine.
        :rtype: str or None, bool or None
        """
        bundle_temp_filename = fetch_file(module=self.module, url=bundle_uri)
        bundle_version = None
        is_multi_tenant = None
        gen = None

        # If not tarfile, then if the file has "MMG2" or "DPG2" at 2048th byte
        # then the bundle is for MM or DP G2
        if not tarfile.is_tarfile(bundle_temp_filename):
            cookie1 = None
            with open(bundle_temp_filename, "rb") as bundle_file:
                file_size = os.path.getsize(bundle_temp_filename)
                if file_size >= 2052:
                    bundle_file.seek(2048)
                    cookie1 = bundle_file.read(4)
            # It is anticipated that DP firmware bundle will be having the value "DPG2"
            # for cookie1 in the header
            if cookie1 and cookie1.decode("utf8") == "MMG2" or cookie1.decode("utf8") == "DPG2":
                file_name, ext = os.path.splitext(str(bundle_uri.rsplit("/", 1)[1]))
                # G2 bundle file name: Ultrastar-Data102_3000_SEP_1010-032_2.1.12
                parsedFileName = file_name.split("_")
                if len(parsedFileName) == 5:
                    bundle_version = parsedFileName[4]
                    # MM G2 is always single tanant
                    is_multi_tenant = False
                    gen = "G2"

            return bundle_version, is_multi_tenant, gen

        # Bundle is for MM or DP G1
        tf = tarfile.open(bundle_temp_filename)
        pattern_pkg = r"oobm-(.+)\.pkg"
        pattern_bin = r"(.*\.bin)"
        bundle_version = None
        is_multi_tenant = None
        for filename in tf.getnames():
            match_pkg = re.match(pattern_pkg, filename)
            if match_pkg is not None:
                bundle_version = match_pkg.group(1)
            match_bin = re.match(pattern_bin, filename)
            if match_bin is not None:
                bin_filename = match_bin.group(1)
                bin_file = tf.extractfile(bin_filename)
                bin_file.seek(11)
                byte_11 = bin_file.read(1)
                is_multi_tenant = byte_11 == b"\x80"
                gen = "G1"

        return bundle_version, is_multi_tenant, gen

    @staticmethod
    def uri_is_http(uri):
        """Return True if the specified URI is http or https.

        :param str uri: A URI.
        :return: True if the URI is http or https, else False
        :rtype: bool
        """
        parsed_bundle_uri = urlparse(uri)
        return parsed_bundle_uri.scheme.lower() in ["http", "https"]

    def update_and_activate(self, update_opts):
        """Update and activate the firmware in a single action.

        Orchestrates the firmware update so that everything can be done in a single command.
        Compares the update version with the already-installed version -- skips update if they are the same.
        Performs retries, handles timeouts as needed.

        """
        # Convert credentials to standard HTTP format
        if (
            update_opts.get("update_creds") is not None
            and "username" in update_opts["update_creds"]
            and "password" in update_opts["update_creds"]
        ):
            update_creds = update_opts["update_creds"]
            parsed_url = urlparse(update_opts["update_image_uri"])
            if update_creds:
                original_netloc = parsed_url.netloc
                parsed_url = parsed_url._replace(
                    netloc=f"{update_creds.get('username')}:{update_creds.get('password')}@{original_netloc}"
                )
                update_opts["update_image_uri"] = urlunparse(parsed_url)
                del update_opts["update_creds"]

        # Make sure bundle URI is HTTP(s)
        bundle_uri = update_opts["update_image_uri"]

        if not self.uri_is_http(bundle_uri):
            return {"ret": False, "msg": "Bundle URI must be HTTP or HTTPS"}
        # Make sure IOM is ready for update
        result = self.get_simple_update_status()
        if result["ret"] is False:
            return result
        update_status = result["entries"]
        status_code = update_status["StatusCode"]
        status_description = update_status["Description"]
        if status_code not in [self.UPDATE_STATUS_CODE_READY_FOR_FW_UPDATE, self.UPDATE_STATUS_CODE_FW_UPDATE_FAILED]:
            return {
                "ret": False,
                "msg": f"Target is not ready for FW update.  Current status: {status_code} ({status_description})",
            }

        # Check the FW version in the bundle file, and compare it to what is already on the IOMs

        # Bundle version number
        bundle_firmware_version, is_bundle_multi_tenant, bundle_gen = self._get_bundle_version(bundle_uri)
        if bundle_firmware_version is None or is_bundle_multi_tenant is None or bundle_gen is None:
            return {
                "ret": False,
                "msg": "Unable to extract bundle version or multi-tenant status or generation from update image file",
            }

        is_enclosure_multi_tenant, enclosure_gen = self._is_enclosure_multi_tenant_and_fetch_gen()

        # Verify that the bundle is correctly multi-tenant or not
        if is_enclosure_multi_tenant != is_bundle_multi_tenant:
            return {
                "ret": False,
                "msg": f"Enclosure multi-tenant is {is_enclosure_multi_tenant} but bundle multi-tenant is {is_bundle_multi_tenant}",
            }

        # Verify that the bundle is compliant with the target enclosure
        if enclosure_gen != bundle_gen:
            return {"ret": False, "msg": f"Enclosure generation is {enclosure_gen} but bundle is of {bundle_gen}"}

        # Version number installed on IOMs
        firmware_inventory = self.get_firmware_inventory()
        if not firmware_inventory["ret"]:
            return firmware_inventory
        firmware_inventory_dict = {}
        for entry in firmware_inventory["entries"]:
            firmware_inventory_dict[entry["Id"]] = entry
        iom_a_firmware_version = firmware_inventory_dict.get("IOModuleA_OOBM", {}).get("Version")
        iom_b_firmware_version = firmware_inventory_dict.get("IOModuleB_OOBM", {}).get("Version")
        # If version is None, we will proceed with the update, because we cannot tell
        # for sure that we have a full version match.
        if is_enclosure_multi_tenant:
            # For multi-tenant, only one of the IOMs will be affected by the firmware update,
            # so see if that IOM already has the same firmware version as the bundle.
            firmware_already_installed = (
                bundle_firmware_version
                == self._get_installed_firmware_version_of_multi_tenant_system(
                    iom_a_firmware_version, iom_b_firmware_version
                )
            )
        else:
            # For single-tenant, see if both IOMs already have the same firmware version as the bundle.
            firmware_already_installed = bundle_firmware_version == iom_a_firmware_version == iom_b_firmware_version
        # If this FW already installed, return changed: False, and do not update the firmware.
        if firmware_already_installed:
            return {"ret": True, "changed": False, "msg": f"Version {bundle_firmware_version} already installed"}

        # Version numbers don't match the bundle -- proceed with update (unless we are in check mode)
        if self.module.check_mode:
            return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}
        update_successful = False
        retry_interval_seconds = 5
        max_number_of_retries = 5
        retry_number = 0
        while retry_number < max_number_of_retries and not update_successful:
            if retry_number != 0:
                time.sleep(retry_interval_seconds)
            retry_number += 1

            result = self.simple_update(update_opts)
            if result["ret"] is not True:
                # Sometimes a timeout error is returned even though the update actually was requested.
                # Check the update status to see if the update is in progress.
                status_result = self.get_simple_update_status()
                if status_result["ret"] is False:
                    continue
                update_status = status_result["entries"]
                status_code = update_status["StatusCode"]
                if status_code != self.UPDATE_STATUS_CODE_FW_UPDATE_IN_PROGRESS:
                    # Update is not in progress -- retry until max number of retries
                    continue
                else:
                    update_successful = True
            else:
                update_successful = True
        if not update_successful:
            # Unable to get SimpleUpdate to work.  Return the failure from the SimpleUpdate
            return result

        # Wait for "ready to activate"
        max_wait_minutes = 30
        polling_interval_seconds = 30
        status_code = self.UPDATE_STATUS_CODE_READY_FOR_FW_UPDATE
        start_time = datetime.datetime.now()
        # For a short time, target will still say "ready for firmware update" before it transitions
        # to "update in progress"
        status_codes_for_update_incomplete = [
            self.UPDATE_STATUS_CODE_FW_UPDATE_IN_PROGRESS,
            self.UPDATE_STATUS_CODE_READY_FOR_FW_UPDATE,
        ]
        iteration = 0
        while (
            status_code in status_codes_for_update_incomplete
            and datetime.datetime.now() - start_time < datetime.timedelta(minutes=max_wait_minutes)
        ):
            if iteration != 0:
                time.sleep(polling_interval_seconds)
            iteration += 1
            result = self.get_simple_update_status()
            if result["ret"] is False:
                continue  # We may get timeouts, just keep trying until we give up
            update_status = result["entries"]
            status_code = update_status["StatusCode"]
            status_description = update_status["Description"]
            if status_code == self.UPDATE_STATUS_CODE_FW_UPDATE_IN_PROGRESS:
                # Once it says update in progress, "ready for update" is no longer a valid status code
                status_codes_for_update_incomplete = [self.UPDATE_STATUS_CODE_FW_UPDATE_IN_PROGRESS]

        # Update no longer in progress -- verify that it finished
        if status_code != self.UPDATE_STATUS_CODE_FW_UPDATE_COMPLETED_WAITING_FOR_ACTIVATION:
            return {
                "ret": False,
                "msg": f"Target is not ready for FW activation after update.  Current status: {status_code} ({status_description})",
            }

        self.firmware_activate(update_opts)
        return {"ret": True, "changed": True, "msg": "Firmware updated and activation initiated."}

    def _get_installed_firmware_version_of_multi_tenant_system(self, iom_a_firmware_version, iom_b_firmware_version):
        """Return the version for the active IOM on a multi-tenant system.

        Only call this on a multi-tenant system.
        Given the installed firmware versions for IOM A, B, this method will determine which IOM is active
        for this tenanat, and return that IOM's firmware version.
        """
        # To determine which IOM we are on, try to GET each IOM resource
        # The one we are on will return valid data.
        # The other will return an error with message "IOM Module A/B cannot be read"
        which_iom_is_this = None
        for iom_letter in ["A", "B"]:
            iom_uri = f"Chassis/IOModule{iom_letter}FRU"
            response = self.get_request(self.root_uri + self.service_root + iom_uri)
            if response["ret"] is False:
                continue
            data = response["data"]
            if "Id" in data:  # Assume if there is an "Id", it is valid
                which_iom_is_this = iom_letter
                break
        if which_iom_is_this == "A":
            return iom_a_firmware_version
        elif which_iom_is_this == "B":
            return iom_b_firmware_version
        else:
            return None

    @staticmethod
    def _get_led_locate_uri(data):
        """Get the LED locate URI given a resource body."""
        if WdcRedfishUtils.ACTIONS not in data:
            return None
        if WdcRedfishUtils.OEM not in data[WdcRedfishUtils.ACTIONS]:
            return None
        if WdcRedfishUtils.WDC not in data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM]:
            return None
        if (
            WdcRedfishUtils.CHASSIS_LOCATE
            not in data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM][WdcRedfishUtils.WDC]
        ):
            return None
        if (
            WdcRedfishUtils.TARGET
            not in data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM][WdcRedfishUtils.WDC][
                WdcRedfishUtils.CHASSIS_LOCATE
            ]
        ):
            return None
        return data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM][WdcRedfishUtils.WDC][WdcRedfishUtils.CHASSIS_LOCATE][
            WdcRedfishUtils.TARGET
        ]

    @staticmethod
    def _get_power_mode_uri(data):
        """Get the Power Mode URI given a resource body."""
        if WdcRedfishUtils.ACTIONS not in data:
            return None
        if WdcRedfishUtils.OEM not in data[WdcRedfishUtils.ACTIONS]:
            return None
        if WdcRedfishUtils.WDC not in data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM]:
            return None
        if (
            WdcRedfishUtils.CHASSIS_POWER_MODE
            not in data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM][WdcRedfishUtils.WDC]
        ):
            return None
        if (
            WdcRedfishUtils.TARGET
            not in data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM][WdcRedfishUtils.WDC][
                WdcRedfishUtils.CHASSIS_POWER_MODE
            ]
        ):
            return None
        return data[WdcRedfishUtils.ACTIONS][WdcRedfishUtils.OEM][WdcRedfishUtils.WDC][
            WdcRedfishUtils.CHASSIS_POWER_MODE
        ][WdcRedfishUtils.TARGET]

    def manage_indicator_led(self, command, resource_uri):
        key = "IndicatorLED"

        payloads = {"IndicatorLedOn": "On", "IndicatorLedOff": "Off"}
        current_led_status_map = {"IndicatorLedOn": "Blinking", "IndicatorLedOff": "Off"}

        result = {}
        response = self.get_request(self.root_uri + resource_uri)
        if response["ret"] is False:
            return response
        result["ret"] = True
        data = response["data"]
        if key not in data:
            return {"ret": False, "msg": f"Key {key} not found"}
        current_led_status = data[key]
        if current_led_status == current_led_status_map[command]:
            return {"ret": True, "changed": False}

        led_locate_uri = self._get_led_locate_uri(data)
        if led_locate_uri is None:
            return {"ret": False, "msg": "LED locate URI not found."}

        if command in payloads.keys():
            payload = {"LocateState": payloads[command]}
            response = self.post_request(self.root_uri + led_locate_uri, payload)
            if response["ret"] is False:
                return response
        else:
            return {"ret": False, "msg": "Invalid command"}

        return result

    def manage_chassis_power_mode(self, command):
        return self.manage_power_mode(command, self.chassis_uri)

    def manage_power_mode(self, command, resource_uri=None):
        if resource_uri is None:
            resource_uri = self.chassis_uri

        payloads = {"PowerModeNormal": "Normal", "PowerModeLow": "Low"}
        requested_power_mode = payloads[command]

        result = {}
        response = self.get_request(self.root_uri + resource_uri)
        if response["ret"] is False:
            return response
        result["ret"] = True
        data = response["data"]

        # Make sure the response includes Oem.WDC.PowerMode, and get current power mode
        power_mode = "PowerMode"
        if (
            WdcRedfishUtils.OEM not in data
            or WdcRedfishUtils.WDC not in data[WdcRedfishUtils.OEM]
            or power_mode not in data[WdcRedfishUtils.OEM][WdcRedfishUtils.WDC]
        ):
            return {"ret": False, "msg": "Resource does not support Oem.WDC.PowerMode"}
        current_power_mode = data[WdcRedfishUtils.OEM][WdcRedfishUtils.WDC][power_mode]
        if current_power_mode == requested_power_mode:
            return {"ret": True, "changed": False}

        power_mode_uri = self._get_power_mode_uri(data)
        if power_mode_uri is None:
            return {"ret": False, "msg": "Power Mode URI not found."}

        if command in payloads.keys():
            payload = {"PowerMode": payloads[command]}
            response = self.post_request(self.root_uri + power_mode_uri, payload)
            if response["ret"] is False:
                return response
        else:
            return {"ret": False, "msg": "Invalid command"}

        return result



#!/usr/bin/python

# Copyright (c) 2017 John Kwiatkoski (@JayKayy) <jkwiat40@gmail.com>
# Copyright (c) 2018 Alexander Bethke (@oolongbrothers) <oolongbrothers@gmx.net>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: flatpak_remote
short_description: Manage flatpak repository remotes
description:
  - Allows users to add or remove flatpak remotes.
  - The flatpak remotes concept is comparable to what is called repositories in other packaging formats.
  - Currently, remote addition is only supported using C(flatpakrepo) file URLs.
  - Existing remotes are not updated.
  - See the M(community.general.flatpak) module for managing flatpaks.
author:
  - John Kwiatkoski (@JayKayy)
  - Alexander Bethke (@oolongbrothers)
requirements:
  - flatpak
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  executable:
    description:
      - The path to the C(flatpak) executable to use.
      - By default, this module looks for the C(flatpak) executable on the path.
    type: str
    default: flatpak
  flatpakrepo_url:
    description:
      - The URL to the C(flatpakrepo) file representing the repository remote to add.
      - When used with O(state=present), the flatpak remote specified under the O(flatpakrepo_url) is added using the specified
        installation O(method).
      - When used with O(state=absent), this is not required.
      - Required when O(state=present).
    type: str
  method:
    description:
      - The installation method to use.
      - Defines if the C(flatpak) is supposed to be installed globally for the whole V(system) or only for the current V(user).
    type: str
    choices: [system, user]
    default: system
  name:
    description:
      - The desired name for the flatpak remote to be registered under on the managed host.
      - When used with O(state=present), the remote is added to the managed host under the specified O(name).
      - When used with O(state=absent) the remote with that name is removed.
    type: str
    required: true
  state:
    description:
      - Indicates the desired package state.
    type: str
    choices: [absent, present]
    default: present
  enabled:
    description:
      - Indicates whether this remote is enabled.
    type: bool
    default: true
    version_added: 6.4.0
"""

EXAMPLES = r"""
- name: Add the Gnome flatpak remote to the system installation
  community.general.flatpak_remote:
    name: gnome
    state: present
    flatpakrepo_url: https://sdk.gnome.org/gnome-apps.flatpakrepo

- name: Add the flathub flatpak repository remote to the user installation
  community.general.flatpak_remote:
    name: flathub
    state: present
    flatpakrepo_url: https://dl.flathub.org/repo/flathub.flatpakrepo
    method: user

- name: Remove the Gnome flatpak remote from the user installation
  community.general.flatpak_remote:
    name: gnome
    state: absent
    method: user

- name: Remove the flathub remote from the system installation
  community.general.flatpak_remote:
    name: flathub
    state: absent

- name: Disable the flathub remote in the system installation
  community.general.flatpak_remote:
    name: flathub
    state: present
    enabled: false
"""

RETURN = r"""
command:
  description: The exact flatpak command that was executed.
  returned: When a flatpak command has been executed
  type: str
  sample: "/usr/bin/flatpak remote-add --system flatpak-test https://dl.flathub.org/repo/flathub.flatpakrepo"
"""

from ansible.module_utils.basic import AnsibleModule


def add_remote(module, binary, name, flatpakrepo_url, method):
    """Add a new remote."""
    global result  # pylint: disable=global-variable-not-assigned
    command = [binary, "remote-add", f"--{method}", name, flatpakrepo_url]
    _flatpak_command(module, module.check_mode, command)
    result["changed"] = True


def remove_remote(module, binary, name, method):
    """Remove an existing remote."""
    global result  # pylint: disable=global-variable-not-assigned
    command = [binary, "remote-delete", f"--{method}", "--force", name]
    _flatpak_command(module, module.check_mode, command)
    result["changed"] = True


def remote_exists(module, binary, name, method):
    """Check if the remote exists."""
    command = [binary, "remote-list", "--show-disabled", f"--{method}"]
    # The query operation for the remote needs to be run even in check mode
    output = _flatpak_command(module, False, command)
    for line in output.splitlines():
        listed_remote = line.split()
        if len(listed_remote) == 0:
            continue
        if listed_remote[0] == name:
            return True
    return False


def enable_remote(module, binary, name, method):
    """Enable a remote."""
    global result  # pylint: disable=global-variable-not-assigned
    command = [binary, "remote-modify", "--enable", f"--{method}", name]
    _flatpak_command(module, module.check_mode, command)
    result["changed"] = True


def disable_remote(module, binary, name, method):
    """Disable a remote."""
    global result  # pylint: disable=global-variable-not-assigned
    command = [binary, "remote-modify", "--disable", f"--{method}", name]
    _flatpak_command(module, module.check_mode, command)
    result["changed"] = True


def remote_enabled(module, binary, name, method):
    """Check if the remote is enabled."""
    command = [binary, "remote-list", "--show-disabled", f"--{method}"]
    # The query operation for the remote needs to be run even in check mode
    output = _flatpak_command(module, False, command)
    for line in output.splitlines():
        listed_remote = line.split()
        if len(listed_remote) == 0:
            continue
        if listed_remote[0] == name:
            return len(listed_remote) == 1 or "disabled" not in listed_remote[1].split(",")
    return False


def _flatpak_command(module, noop, command):
    global result  # pylint: disable=global-variable-not-assigned
    result["command"] = " ".join(command)
    if noop:
        result["rc"] = 0
        return ""

    result["rc"], result["stdout"], result["stderr"] = module.run_command(command, check_rc=True)
    return result["stdout"]


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            flatpakrepo_url=dict(type="str"),
            method=dict(type="str", default="system", choices=["user", "system"]),
            state=dict(type="str", default="present", choices=["absent", "present"]),
            enabled=dict(type="bool", default=True),
            executable=dict(type="str", default="flatpak"),
        ),
        # This module supports check mode
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    name = module.params["name"]
    flatpakrepo_url = module.params["flatpakrepo_url"]
    method = module.params["method"]
    state = module.params["state"]
    enabled = module.params["enabled"]
    executable = module.params["executable"]
    binary = module.get_bin_path(executable, None)

    if flatpakrepo_url is None:
        flatpakrepo_url = ""

    global result
    result = dict(changed=False)

    # If the binary was not found, fail the operation
    if not binary:
        module.fail_json(msg=f"Executable '{executable}' was not found on the system.", **result)

    remote_already_exists = remote_exists(module, binary, name, method)

    if state == "present" and not remote_already_exists:
        add_remote(module, binary, name, flatpakrepo_url, method)
    elif state == "absent" and remote_already_exists:
        remove_remote(module, binary, name, method)

    if state == "present":
        remote_already_enabled = remote_enabled(module, binary, name, method)

        if enabled and not remote_already_enabled:
            enable_remote(module, binary, name, method)
        if not enabled and remote_already_enabled:
            disable_remote(module, binary, name, method)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_saml_idps
short_description: Resource module for organizations _saml _idps
description:
  - Manage operations create, update and delete of the resource organizations _saml
    _idps.
  - Create a SAML IdP for your organization.
  - Remove a SAML IdP in your organization.
  - Update a SAML IdP in your organization.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  idpId:
    description: IdpId path parameter. Idp ID.
    type: str
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  sloLogoutUrl:
    description: Dashboard will redirect users to this URL when they sign out.
    type: str
  ssoLoginUrl:
    description: Dashboard will redirect users to this URL to log in again when their
      sessions expire.
    type: str
  x509certSha1Fingerprint:
    description: Fingerprint (SHA1) of the SAML certificate provided by your Identity
      Provider (IdP). This will be used for encryption / validation.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations createOrganizationSamlIdp
    description: Complete reference of the createOrganizationSamlIdp API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-saml-idp
  - name: Cisco Meraki documentation for organizations deleteOrganizationSamlIdp
    description: Complete reference of the deleteOrganizationSamlIdp API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-organization-saml-idp
  - name: Cisco Meraki documentation for organizations updateOrganizationSamlIdp
    description: Complete reference of the updateOrganizationSamlIdp API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-saml-idp
notes:
  - SDK Method used are
    organizations.Organizations.create_organization_saml_idp,
    organizations.Organizations.delete_organization_saml_idp,
    organizations.Organizations.update_organization_saml_idp,
  - Paths used are
    post /organizations/{organizationId}/saml/idps,
    delete /organizations/{organizationId}/saml/idps/{idpId},
    put /organizations/{organizationId}/saml/idps/{idpId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_saml_idps:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    organizationId: string
    sloLogoutUrl: https://onelogin.com/trust/saml2/http-redirect/slo/4155000
    ssoLoginUrl: https://onelogin.com/trust/saml2/http-post/sso/3de5f942-e7b8-4cb9-94e3-85828111158b
    x509certSha1Fingerprint: 00:11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99
- name: Update by id
  cisco.meraki.organizations_saml_idps:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    idpId: string
    organizationId: string
    sloLogoutUrl: https://onelogin.com/trust/saml2/http-redirect/slo/4155000
    ssoLoginUrl: https://onelogin.com/trust/saml2/http-post/sso/3de5f942-e7b8-4cb9-94e3-85828111158b
    x509certSha1Fingerprint: 00:11:22:33:44:55:66:77:88:99:00:11:22:33:44:55:66:77:88:99
- name: Delete by id
  cisco.meraki.organizations_saml_idps:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    idpId: string
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  sample: >
    [
      {
        "idpId": "string",
        "consumerUrl": "string",
        "visionConsumerUrl": "string",
        "x509certSha1Fingerprint": "string",
        "ssoLoginUrl": "string",
        "sloLogoutUrl": "string"
      }
    ]
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type
from ansible.plugins.action import ActionBase
try:
    from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
        AnsibleArgSpecValidator, )
except ImportError:
    ANSIBLE_UTILS_IS_INSTALLED = False
else:
    ANSIBLE_UTILS_IS_INSTALLED = True
from ansible.errors import AnsibleActionFail
from ansible_collections.cisco.meraki.plugins.plugin_utils.meraki import (
    MERAKI,
    meraki_argument_spec,
    meraki_compare_equality2,
    get_dict_result,
)
from ansible_collections.cisco.meraki.plugins.plugin_utils.exceptions import (
    InconsistentParameters,
)

# Get common arguments specification
argument_spec = meraki_argument_spec()
# Add arguments specific for this module
argument_spec.update(dict(
    state=dict(type="str", default="present", choices=["present", "absent"]),
    name=dict(type="str"),
    center=dict(type="dict"),
    bottomLeftCorner=dict(type="dict"),
    bottomRightCorner=dict(type="dict"),
    topLeftCorner=dict(type="dict"),
    topRightCorner=dict(type="dict"),
    floorNumber=dict(type="float"),
    imageContents=dict(type="str"),
    networkId=dict(type="str"),
    floorPlanId=dict(type="str"),
))

required_if = [
    ("state", "present", ["floorPlanId", "name", "networkId"], True),
    ("state", "absent", ["floorPlanId", "name", "networkId"], True),
]
required_one_of = []
mutually_exclusive = []
required_together = []


class NetworksFloorPlans(object):
    def __init__(self, params, meraki):
        self.meraki = meraki
        self.new_object = dict(
            name=params.get("name"),
            center=params.get("center"),
            bottomLeftCorner=params.get("bottomLeftCorner"),
            bottomRightCorner=params.get("bottomRightCorner"),
            topLeftCorner=params.get("topLeftCorner"),
            topRightCorner=params.get("topRightCorner"),
            floorNumber=params.get("floorNumber"),
            imageContents=params.get("imageContents"),
            networkId=params.get("networkId"),
            floorPlanId=params.get("floorPlanId"),
        )

    def get_all_params(self, name=None, id=None):
        new_object_params = {}
        if self.new_object.get('networkId') is not None or self.new_object.get(
                'network_id') is not None:
            new_object_params['networkId'] = self.new_object.get(
                'networkId') or self.new_object.get('network_id')
        return new_object_params

    def get_params_by_id(self, name=None, id=None):
        new_object_params = {}
        if self.new_object.get('networkId') is not None or self.new_object.get(
                'network_id') is not None:
            new_object_params['networkId'] = self.new_object.get(
                'networkId') or self.new_object.get('network_id')
        if self.new_object.get('floorPlanId') is not None or self.new_object.get(
                'floor_plan_id') is not None:
            new_object_params['floorPlanId'] = self.new_object.get(
                'floorPlanId') or self.new_object.get('floor_plan_id')
        return new_object_params

    def create_params(self):
        new_object_params = {}
        if self.new_object.get('name') is not None or self.new_object.get(
                'name') is not None:
            new_object_params['name'] = self.new_object.get('name') or \
                self.new_object.get('name')
        if self.new_object.get('center') is not None or self.new_object.get(
                'center') is not None:
            new_object_params['center'] = self.new_object.get('center') or \
                self.new_object.get('center')
        if self.new_object.get('bottomLeftCorner') is not None or self.new_object.get(
                'bottom_left_corner') is not None:
            new_object_params['bottomLeftCorner'] = self.new_object.get(
                'bottomLeftCorner') or self.new_object.get('bottom_left_corner')
        if self.new_object.get('bottomRightCorner') is not None or self.new_object.get(
                'bottom_right_corner') is not None:
            new_object_params['bottomRightCorner'] = self.new_object.get(
                'bottomRightCorner') or self.new_object.get('bottom_right_corner')
        if self.new_object.get('topLeftCorner') is not None or self.new_object.get(
                'top_left_corner') is not None:
            new_object_params['topLeftCorner'] = self.new_object.get(
                'topLeftCorner') or self.new_object.get('top_left_corner')
        if self.new_object.get('topRightCorner') is not None or self.new_object.get(
                'top_right_corner') is not None:
            new_object_params['topRightCorner'] = self.new_object.get(
                'topRightCorner') or self.new_object.get('top_right_corner')
        if self.new_object.get('floorNumber') is not None or self.new_object.get(
                'floor_number') is not None:
            new_object_params['floorNumber'] = self.new_object.get(
                'floorNumber') or self.new_object.get('floor_number')
        if self.new_object.get('imageContents') is not None or self.new_object.get(
                'image_contents') is not None:
            new_object_params['imageContents'] = self.new_object.get(
                'imageContents') or self.new_object.get('image_contents')
        if self.new_object.get('networkId') is not None or self.new_object.get(
                'network_id') is not None:
            new_object_params['networkId'] = self.new_object.get(
                'networkId') or self.new_object.get('network_id')
        return new_object_params

    def delete_by_id_params(self):
        new_object_params = {}
        if self.new_object.get('networkId') is not None or self.new_object.get(
                'network_id') is not None:
            new_object_params['networkId'] = self.new_object.get(
                'networkId') or self.new_object.get('network_id')
        if self.new_object.get('floorPlanId') is not None or self.new_object.get(
                'floor_plan_id') is not None:
            new_object_params['floorPlanId'] = self.new_object.get(
                'floorPlanId') or self.new_object.get('floor_plan_id')
        return new_object_params

    def update_by_id_params(self):
        new_object_params = {}
        if self.new_object.get('name') is not None or self.new_object.get(
                'name') is not None:
            new_object_params['name'] = self.new_object.get('name') or \
                self.new_object.get('name')
        if self.new_object.get('center') is not None or self.new_object.get(
                'center') is not None:
            new_object_params['center'] = self.new_object.get('center') or \
                self.new_object.get('center')
        if self.new_object.get('bottomLeftCorner') is not None or self.new_object.get(
                'bottom_left_corner') is not None:
            new_object_params['bottomLeftCorner'] = self.new_object.get(
                'bottomLeftCorner') or self.new_object.get('bottom_left_corner')
        if self.new_object.get('bottomRightCorner') is not None or self.new_object.get(
                'bottom_right_corner') is not None:
            new_object_params['bottomRightCorner'] = self.new_object.get(
                'bottomRightCorner') or self.new_object.get('bottom_right_corner')
        if self.new_object.get('topLeftCorner') is not None or self.new_object.get(
                'top_left_corner') is not None:
            new_object_params['topLeftCorner'] = self.new_object.get(
                'topLeftCorner') or self.new_object.get('top_left_corner')
        if self.new_object.get('topRightCorner') is not None or self.new_object.get(
                'top_right_corner') is not None:
            new_object_params['topRightCorner'] = self.new_object.get(
                'topRightCorner') or self.new_object.get('top_right_corner')
        if self.new_object.get('floorNumber') is not None or self.new_object.get(
                'floor_number') is not None:
            new_object_params['floorNumber'] = self.new_object.get(
                'floorNumber') or self.new_object.get('floor_number')
        if self.new_object.get('imageContents') is not None or self.new_object.get(
                'image_contents') is not None:
            new_object_params['imageContents'] = self.new_object.get(
                'imageContents') or self.new_object.get('image_contents')
        if self.new_object.get('networkId') is not None or self.new_object.get(
                'network_id') is not None:
            new_object_params['networkId'] = self.new_object.get(
                'networkId') or self.new_object.get('network_id')
        if self.new_object.get('floorPlanId') is not None or self.new_object.get(
                'floor_plan_id') is not None:
            new_object_params['floorPlanId'] = self.new_object.get(
                'floorPlanId') or self.new_object.get('floor_plan_id')
        return new_object_params

    def get_object_by_name(self, name):
        result = None
        # NOTE: Does not have a get by name method or it is in another action
        try:
            items = self.meraki.exec_meraki(
                family="networks",
                function="getNetworkFloorPlans",
                params=self.get_all_params(name=name),
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = get_dict_result(items, 'name', name)
            if result is None:
                result = items
        except Exception as e:
            print("Error: ", e)
            result = None
        return result

    def get_object_by_id(self, id):
        result = None
        try:
            items = self.meraki.exec_meraki(
                family="networks",
                function="getNetworkFloorPlan",
                params=self.get_params_by_id()
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = get_dict_result(items, 'floorPlanId', id)
        except Exception as e:
            print("Error: ", e)
            result = None
        return result

    def exists(self):
        id_exists = False
        name_exists = False
        prev_obj = None
        o_id = self.new_object.get("id")
        o_id = o_id or self.new_object.get(
            "floor_plan_id") or self.new_object.get("floorPlanId")
        name = self.new_object.get("name")
        if o_id:
            prev_obj = self.get_object_by_id(o_id)
            id_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if not id_exists and name:
            prev_obj = self.get_object_by_name(name)
            name_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if name_exists:
            _id = prev_obj.get("id")
            _id = _id or prev_obj.get("floorPlanId")
            if id_exists and name_exists and o_id != _id:
                raise InconsistentParameters(
                    "The 'id' and 'name' params don't refer to the same object")
            if _id:
                self.new_object.update(dict(id=_id))
                self.new_object.update(dict(floorPlanId=_id))
            if _id:
                prev_obj = self.get_object_by_id(_id)
        it_exists = prev_obj is not None and isinstance(prev_obj, dict)
        return (it_exists, prev_obj)

    def requires_update(self, current_obj):
        requested_obj = self.new_object

        obj_params = [
            ("name", "name"),
            ("center", "center"),
            ("bottomLeftCorner", "bottomLeftCorner"),
            ("bottomRightCorner", "bottomRightCorner"),
            ("topLeftCorner", "topLeftCorner"),
            ("topRightCorner", "topRightCorner"),
            ("floorNumber", "floorNumber"),
            ("imageContents", "imageContents"), ("floorPlanId", "floorPlanId"),
        ]
        # Method 1. Params present in request (Ansible) obj are the same as the current (DNAC) params
        # If any does not have eq params, it requires update
        return any(
            not meraki_compare_equality2(
                current_obj.get(meraki_param),
                requested_obj.get(ansible_param)) for (
                meraki_param,
                ansible_param) in obj_params)

    def create(self):
        result = self.meraki.exec_meraki(
            family="networks",
            function="createNetworkFloorPlan",
            params=self.create_params(),
            op_modifies=True,
        )
        return result

    def update(self):
        id = self.new_object.get("id")
        id = id or self.new_object.get("floorPlanId")
        name = self.new_object.get("name")
        result = None
        if not id:
            prev_obj_name = self.get_object_by_name(name)
            id_ = None
            if prev_obj_name:
                id_ = prev_obj_name.get("id")
                id_ = id_ or prev_obj_name.get("floorPlanId")
            if id_:
                self.new_object.update(dict(floorPlanId=id_))
        result = self.meraki.exec_meraki(
            family="networks",
            function="updateNetworkFloorPlan",
            params=self.update_by_id_params(),
            op_modifies=True,
        )
        return result

    def delete(self):
        id = self.new_object.get("id")
        id = id or self.new_object.get("floorPlanId")
        name = self.new_object.get("name")
        result = None
        if not id:
            prev_obj_name = self.get_object_by_name(name)
            id_ = None
            if prev_obj_name:
                id_ = prev_obj_name.get("id")
                id_ = id_ or prev_obj_name.get("floorPlanId")
            if id_:
                self.new_object.update(dict(floorPlanId=id_))
        result = self.meraki.exec_meraki(
            family="networks",
            function="deleteNetworkFloorPlan",
            params=self.delete_by_id_params(),
        )
        return result


class ActionModule(ActionBase):
    def __init__(self, *args, **kwargs):
        if not ANSIBLE_UTILS_IS_INSTALLED:
            raise AnsibleActionFail(
                "ansible.utils is not installed. Execute 'ansible-galaxy collection install ansible.utils'")
        super(ActionModule, self).__init__(*args, **kwargs)
        self._supports_async = False
        self._supports_check_mode = False
        self._result = None

    # Checks the supplied parameters against the argument spec for this module
    def _check_argspec(self):
        aav = AnsibleArgSpecValidator(
            data=self._task.args,
            schema=dict(argument_spec=argument_spec),
            schema_format="argspec",
            schema_conditionals=dict(
                required_if=required_if,
                required_one_of=required_one_of,
                mutually_exclusive=mutually_exclusive,
                required_together=required_together,
            ),
            name=self._task.action,
        )
        valid, errors, self._task.args = aav.validate()
        if not valid:
            raise AnsibleActionFail(errors)

    def run(self, tmp=None, task_vars=None):
        self._task.diff = False
        self._result = super(ActionModule, self).run(tmp, task_vars)
        self._result["changed"] = False
        self._check_argspec()

        meraki = MERAKI(self._task.args)
        obj = NetworksFloorPlans(self._task.args, meraki)

        state = self._task.args.get("state")

        response = None

        if state == "present":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                if obj.requires_update(prev_obj):
                    response = obj.update()
                    meraki.object_updated()
                else:
                    response = prev_obj
                    meraki.object_already_present()
            else:
                response = obj.create()
                meraki.object_created()

        elif state == "absent":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                response = obj.delete()
                meraki.object_deleted()
            else:
                meraki.object_already_absent()

        self._result.update(dict(meraki_response=response))
        self._result.update(meraki.exit_json())
        return self._result



# -*- coding: utf-8 -*-

# (c) 2023 Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from ansible.errors import AnsibleConnectionFailure
from ansible.plugins.connection import ConnectionBase

from ansible_collections.amazon.aws.plugins.plugin_utils.base import AWSPluginBase


class AWSConnectionBase(AWSPluginBase, ConnectionBase):
    def _do_fail(self, message):
        raise AnsibleConnectionFailure(message)

    def __init__(self, *args, boto3_version=None, botocore_version=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.require_aws_sdk(botocore_version=botocore_version, boto3_version=boto3_version)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Jacob McGill (@jmcgill298)
# Copyright: (c) 2025, Akini Ross (@akinross) <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_bd_subnet
short_description: Manage Subnets (fv:Subnet)
description:
- Manage Subnets on Cisco ACI fabrics.
options:
  bd:
    description:
    - The name of the Bridge Domain.
    type: str
    aliases: [ bd_name ]
  description:
    description:
    - The description for the Subnet.
    type: str
    aliases: [ descr ]
  enable_vip:
    description:
    - Determines if the Subnet should be treated as a VIP; used when the BD is extended to multiple sites.
    - The APIC defaults to C(false) when unset during creation.
    type: bool
  gateway:
    description:
    - The IPv4 or IPv6 gateway address for the Subnet.
    type: str
    aliases: [ gateway_ip ]
  mask:
    description:
    - The subnet mask for the Subnet.
    - This is the number associated with CIDR notation.
    - For IPv4 addresses, accepted values range between C(0) and C(32).
    - For IPv6 addresses, accepted Values range between C(0) and C(128).
    type: int
    aliases: [ subnet_mask ]
  nd_prefix_policy:
    description:
    - The IPv6 Neighbor Discovery Prefix Policy to associate with the Subnet.
    type: str
  preferred:
    description:
    - Determines if the Subnet is preferred over all available Subnets. Only one Subnet per Address Family (IPv4/IPv6).
      can be preferred in the Bridge Domain.
    - The APIC defaults to C(false) when unset during creation.
    type: bool
  route_profile:
    description:
    - The Route Profile to the associate with the Subnet.
    type: str
  route_profile_l3_out:
    description:
    - The L3 Out that contains the associated Route Profile.
    type: str
  scope:
    description:
    - Determines the scope of the Subnet.
    - The C(private) option only allows communication with hosts in the same VRF.
    - The C(public) option allows the Subnet to be advertised outside of the ACI Fabric, and allows communication with
      hosts in other VRFs.
    - The shared option limits communication to hosts in either the same VRF or the shared VRF.
    - The value is a list of options, C(private) and C(public) are mutually exclusive, but both can be used with C(shared).
    - The APIC defaults to C(private) when unset during creation.
    type: list
    elements: str
    choices:
      - private
      - public
      - shared
  subnet_control:
    description:
    - Determines the Subnet's Control State.
    - The C(querier_ip) option is used to treat the gateway_ip as an IGMP querier source IP.
    - The C(nd_ra) option is used to treat the gateway_ip address as a Neighbor Discovery Router Advertisement Prefix.
    - The C(no_gw) option is used to remove default gateway functionality from the gateway address.
    - The APIC defaults to C(nd_ra) when unset during creation.
    type: list
    elements: str
    choices: [ nd_ra, no_gw, querier_ip, unspecified ]
  subnet_name:
    description:
    - The name of the Subnet.
    type: str
    aliases: [ name ]
  ip_data_plane_learning:
    description:
    - Whether IP data plane learning is enabled or disabled.
    - The APIC defaults to C(enabled) when unset during creation.
    type: str
    choices: [ enabled, disabled ]
    aliases: [ ip_dataplane_learning ]
  tenant:
    description:
    - The name of the Tenant.
    type: str
    aliases: [ tenant_name ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
  name_alias:
    description:
    - The alias for the current object. This relates to the nameAlias field in ACI.
    type: str
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

notes:
- The C(gateway) parameter is the root key used to access the Subnet (not name), so the C(gateway)
  is required when the state is C(absent) or C(present).
- The C(tenant) and C(bd) used must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant) module and M(cisco.aci.aci_bd) can be used for these.
seealso:
- module: cisco.aci.aci_bd
- module: cisco.aci.aci_tenant
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(fv:Subnet).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Jacob McGill (@jmcgill298)
- Akini Ross (@akinross)
"""

EXAMPLES = r"""
- name: Create a tenant
  cisco.aci.aci_tenant:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    state: present
  delegate_to: localhost

- name: Create a bridge domain
  cisco.aci.aci_bd:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    bd: database
    state: present
  delegate_to: localhost

- name: Create a subnet
  cisco.aci.aci_bd_subnet:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    bd: database
    gateway: 10.1.1.1
    mask: 24
    state: present
  delegate_to: localhost

- name: Create a subnet with options
  cisco.aci.aci_bd_subnet:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    bd: database
    subnet_name: sql
    gateway: 10.1.2.1
    mask: 23
    description: SQL Servers
    scope: public
    route_profile_l3_out: corp
    route_profile: corp_route_profile
    subnet_control:
      - no_gw
      - querier_ip
    state: present
  delegate_to: localhost

- name: Update a subnets scope to private and shared
  cisco.aci.aci_bd_subnet:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    bd: database
    gateway: 10.1.1.1
    mask: 24
    scope: [private, shared]
    state: present
  delegate_to: localhost

- name: Get all subnets
  cisco.aci.aci_bd_subnet:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost

- name: Get all subnets of specific gateway in specified tenant
  cisco.aci.aci_bd_subnet:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    gateway: 10.1.1.1
    mask: 24
    state: query
  delegate_to: localhost
  register: query_result

- name: Get specific subnet
  cisco.aci.aci_bd_subnet:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    bd: database
    gateway: 10.1.1.1
    mask: 24
    state: query
  delegate_to: localhost
  register: query_result

- name: Delete a subnet
  cisco.aci.aci_bd_subnet:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    bd: database
    gateway: 10.1.1.1
    mask: 24
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

import re
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec
from ansible_collections.cisco.aci.plugins.module_utils.constants import SUBNET_CONTROL_MAPPING_BD_SUBNET, IPV4_REGEX


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        bd=dict(type="str", aliases=["bd_name"]),  # Not required for querying all objects
        description=dict(type="str", aliases=["descr"]),
        enable_vip=dict(type="bool"),
        gateway=dict(type="str", aliases=["gateway_ip"]),  # Not required for querying all objects
        mask=dict(type="int", aliases=["subnet_mask"]),  # Not required for querying all objects
        subnet_name=dict(type="str", aliases=["name"]),
        nd_prefix_policy=dict(type="str"),
        preferred=dict(type="bool"),
        route_profile=dict(type="str"),
        route_profile_l3_out=dict(type="str"),
        scope=dict(type="list", elements="str", choices=["private", "public", "shared"]),
        subnet_control=dict(type="list", elements="str", choices=list(SUBNET_CONTROL_MAPPING_BD_SUBNET)),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        tenant=dict(type="str", aliases=["tenant_name"]),  # Not required for querying all objects
        name_alias=dict(type="str"),
        ip_data_plane_learning=dict(type="str", choices=["enabled", "disabled"], aliases=["ip_dataplane_learning"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_together=[["gateway", "mask"]],
        required_if=[
            ["state", "present", ["bd", "gateway", "mask", "tenant"]],
            ["state", "absent", ["bd", "gateway", "mask", "tenant"]],
        ],
    )

    aci = ACIModule(module)

    description = module.params.get("description")
    enable_vip = aci.boolean(module.params.get("enable_vip"))
    tenant = module.params.get("tenant")
    bd = module.params.get("bd")
    gateway = module.params.get("gateway")
    mask = module.params.get("mask")

    if mask:
        # Assumption for simplicity of code that when a valid IPv4 address is not provided the intend is IPv6.
        if re.search(IPV4_REGEX, gateway) and mask not in range(0, 33):
            module.fail_json(msg="Valid Subnet Masks are 0 to 32 for IPv4 Addresses")
        elif mask not in range(0, 129):
            module.fail_json(msg="Valid Subnet Masks are 0 to 128 for IPv6 Addresses")

    if gateway is not None:
        gateway = "{0}/{1}".format(gateway, str(mask))
    subnet_name = module.params.get("subnet_name")
    nd_prefix_policy = module.params.get("nd_prefix_policy")
    preferred = aci.boolean(module.params.get("preferred"))
    route_profile = module.params.get("route_profile")
    route_profile_l3_out = module.params.get("route_profile_l3_out")
    scope = module.params.get("scope")
    if scope is not None:
        if "private" in scope and "public" in scope:
            module.fail_json(msg="Parameter 'scope' cannot be both 'private' and 'public', got: %s" % scope)
        else:
            scope = ",".join(sorted(scope))
    state = module.params.get("state")
    subnet_control = module.params.get("subnet_control")
    if subnet_control:
        if len(subnet_control) > 1 and "unspecified" in subnet_control:
            module.fail_json(msg="Parameter 'subnet_control' cannot contain 'unspecified' in combination with other values, got: %s" % subnet_control)
        subnet_control = ",".join(sorted([SUBNET_CONTROL_MAPPING_BD_SUBNET[ctrl] for ctrl in subnet_control]))
    name_alias = module.params.get("name_alias")
    ip_data_plane_learning = module.params.get("ip_data_plane_learning")
    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="fvBD",
            aci_rn="BD-{0}".format(bd),
            module_object=bd,
            target_filter={"name": bd},
        ),
        subclass_2=dict(
            aci_class="fvSubnet",
            aci_rn="subnet-[{0}]".format(gateway),
            module_object=gateway,
            target_filter={"ip": gateway},
        ),
        child_classes=["fvRsBDSubnetToProfile", "fvRsNdPfxPol"],
    )

    aci.get_existing()

    if state == "present":
        class_config = dict(
            ctrl=subnet_control,
            descr=description,
            ip=gateway,
            name=subnet_name,
            preferred=preferred,
            scope=scope,
            virtual=enable_vip,
            nameAlias=name_alias,
        )

        if ip_data_plane_learning:
            class_config["ipDPLearning"] = ip_data_plane_learning

        aci.payload(
            aci_class="fvSubnet",
            class_config=class_config,
            child_configs=[
                {"fvRsBDSubnetToProfile": {"attributes": {"tnL3extOutName": route_profile_l3_out, "tnRtctrlProfileName": route_profile}}},
                {"fvRsNdPfxPol": {"attributes": {"tnNdPfxPolName": nd_prefix_policy}}},
            ],
        )

        aci.get_diff(aci_class="fvSubnet")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_ethernet_adapter_policy_info
short_description: Gather information about Ethernet Adapter Policies in Cisco Intersight
description:
  - Gather information about Ethernet Adapter Policies in L(Cisco Intersight,https://intersight.com).
  - Information can be filtered by O(organization) and O(name).
  - If no filters are passed, all Ethernet Adapter Policies will be returned.
  - Returns comprehensive policy configuration including VXLAN, NVGRE, ARFS, PTP, RoCE, queue settings, and offload configurations.
extends_documentation_fragment: intersight
options:
  organization:
    description:
      - The name of the organization the Ethernet Adapter Policy belongs to.
      - Use 'default' for the default organization.
    type: str
  name:
    description:
      - The name of the Ethernet Adapter Policy to gather information from.
    type: str
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Fetch all Ethernet Adapter Policies
  cisco.intersight.intersight_ethernet_adapter_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
  register: all_adapter_policies

- name: Fetch Ethernet Adapter Policies from a specific organization
  cisco.intersight.intersight_ethernet_adapter_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "Engineering"
  register: engineering_adapter_policies

- name: Fetch a specific Ethernet Adapter Policy by name
  cisco.intersight.intersight_ethernet_adapter_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "high-performance-adapter-policy"
  register: specific_adapter_policy
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: list
  sample:
    [
      {
        "Moid": "6889f3e8f654d9f7013baae4",
        "ObjectType": "vnic.EthAdapterPolicy",
        "ClassId": "vnic.EthAdapterPolicy",
        "Name": "high-performance-adapter-policy",
        "Description": "High performance Ethernet adapter policy",
        "AdvancedFilter": true,
        "ArfsSettings": {
          "ClassId": "vnic.ArfsSettings",
          "ObjectType": "vnic.ArfsSettings",
          "Enabled": true
        },
        "CompletionQueueSettings": {
          "ClassId": "vnic.CompletionQueueSettings",
          "ObjectType": "vnic.CompletionQueueSettings",
          "Count": 12,
          "RingSize": 1
        },
        "EtherChannelPinningEnabled": false,
        "GeneveEnabled": false,
        "InterruptScaling": true,
        "InterruptSettings": {
          "ClassId": "vnic.EthInterruptSettings",
          "ObjectType": "vnic.EthInterruptSettings",
          "CoalescingTime": 125,
          "CoalescingType": "MIN",
          "Count": 16,
          "Mode": "MSIx"
        },
        "NvgreSettings": {
          "ClassId": "vnic.NvgreSettings",
          "ObjectType": "vnic.NvgreSettings",
          "Enabled": false
        },
        "PtpSettings": {
          "ClassId": "vnic.PtpSettings",
          "ObjectType": "vnic.PtpSettings",
          "Enabled": true
        },
        "RoceSettings": {
          "ClassId": "vnic.RoceSettings",
          "ObjectType": "vnic.RoceSettings",
          "ClassOfService": 5,
          "Enabled": true,
          "MemoryRegions": 131072,
          "QueuePairs": 256,
          "ResourceGroups": 2,
          "Version": 2
        },
        "RssHashSettings": {
          "ClassId": "vnic.RssHashSettings",
          "ObjectType": "vnic.RssHashSettings",
          "Ipv4Hash": true,
          "Ipv6ExtHash": false,
          "Ipv6Hash": true,
          "TcpIpv4Hash": true,
          "TcpIpv6ExtHash": false,
          "TcpIpv6Hash": true,
          "UdpIpv4Hash": false,
          "UdpIpv6Hash": false
        },
        "RssSettings": true,
        "RxQueueSettings": {
          "ClassId": "vnic.EthRxQueueSettings",
          "ObjectType": "vnic.EthRxQueueSettings",
          "Count": 8,
          "RingSize": 512
        },
        "TcpOffloadSettings": {
          "ClassId": "vnic.TcpOffloadSettings",
          "ObjectType": "vnic.TcpOffloadSettings",
          "LargeReceive": true,
          "LargeSend": true,
          "RxChecksum": true,
          "TxChecksum": true
        },
        "TxQueueSettings": {
          "ClassId": "vnic.EthTxQueueSettings",
          "ObjectType": "vnic.EthTxQueueSettings",
          "Count": 4,
          "RingSize": 256
        },
        "UplinkFailbackTimeout": 5,
        "VxlanSettings": {
          "ClassId": "vnic.VxlanSettings",
          "ObjectType": "vnic.VxlanSettings",
          "Enabled": true
        },
        "Organization": {
          "ObjectType": "organization.Organization",
          "ClassId": "mo.MoRef",
          "Moid": "675450ee69726530014753e2",
          "link": "https://us-east-1.intersight.com/api/v1/organization/Organizations/675450ee69726530014753e2"
        },
        "Tags": [
          {
            "Key": "Environment",
            "Value": "Production"
          },
          {
            "Key": "Feature",
            "Value": "HighPerformance"
          }
        ]
      }
    ]
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        organization=dict(type='str'),
        name=dict(type='str')
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    # Resource path used to fetch info
    resource_path = '/vnic/EthAdapterPolicies'

    query_params = intersight.set_query_params()

    intersight.get_resource(
        resource_path=resource_path,
        query_params=query_params,
        return_list=True
    )

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Andrew Klychkov (@Andersson007) <andrew.a.klychkov@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = r'''
---
module: mysql_query
short_description: Run MySQL or MariaDB queries
description:
- Runs arbitrary MySQL or MariaDB queries.
- Pay attention, the module does not support check mode!
  All queries will be executed in autocommit mode.
- To run SQL queries from a file, use M(ansible.mysql.mysql_db) module.
version_added: '0.1.0'
options:
  query:
    description:
    - SQL query to run. Multiple queries can be passed using YAML list syntax.
    - Must be a string or YAML list containing strings.
    - If you use I(named_args) or I(positional_args) any C(%) will be interpreted
      as a formatting character. All literal C(%) characters in the query should be
      escaped as C(%%).
    - Note that if you use the C(IF EXISTS/IF NOT EXISTS) clauses in your query
      and C(mysqlclient) or C(PyMySQL 0.10.0+) connectors, the module will report
      that the state has been changed even if it has not. If it is important in your
      workflow, use the C(PyMySQL 0.9.3) connector instead.
    type: raw
    required: true
  positional_args:
    description:
    - List of values to be passed as positional arguments to the query.
    - Mutually exclusive with I(named_args).
    type: list
    elements: raw
  named_args:
    description:
    - Dictionary of key-value arguments to pass to the query.
    - Mutually exclusive with I(positional_args).
    type: dict
  login_db:
    description:
    - Name of database to connect to and run queries against.
    type: str
  single_transaction:
    description:
    - Where passed queries run in a single transaction (C(yes)) or commit them one-by-one (C(no)).
    type: bool
    default: false
  session_vars:
     description:
     - "Dictionary of session variables in form of C(variable: value) to set at the beginning of module execution."
     - Cannot be used to set global variables, use the M(ansible.mysql.mysql_variables) module instead.
     type: dict
     version_added: '3.16.0'
attributes:
  check_mode:
    support: none
  idempotent:
    support: none
seealso:
- module: ansible.mysql.mysql_db
notes:
- Compatible with MariaDB or MySQL.
author:
- Andrew Klychkov (@Andersson007)
extends_documentation_fragment:
- ansible.mysql.mysql
'''

EXAMPLES = r'''
# If you encounter the "Please explicitly state intended protocol" error,
# use the login_unix_socket argument
- name: Simple select query to acme db
  ansible.mysql.mysql_query:
    login_db: acme
    query: SELECT * FROM orders
    login_unix_socket: /run/mysqld/mysqld.sock

- name: Select query to db acme with positional arguments
  ansible.mysql.mysql_query:
    login_db: acme
    query: SELECT * FROM acme WHERE id = %s AND story = %s
    positional_args:
      - 1
      - test

- name: Select query to test_db with named_args
  ansible.mysql.mysql_query:
    login_db: test_db
    query: SELECT * FROM test WHERE id = %(id_val)s AND story = %(story_val)s
    named_args:
      id_val: 1
      story_val: test

- name: Run several insert queries against db test_db in single transaction
  ansible.mysql.mysql_query:
    login_db: test_db
    query:
      - INSERT INTO articles (id, story) VALUES (2, 'my_long_story')
      - INSERT INTO prices (id, price) VALUES (123, '100.00')
    single_transaction: true
'''

RETURN = r'''
executed_queries:
    description: List of executed queries.
    returned: always
    type: list
    sample: ['SELECT * FROM bar', 'UPDATE bar SET id = 1 WHERE id = 2']
query_result:
    description:
    - List of lists (sublist for each query) containing dictionaries
      in column:value form representing returned rows.
    returned: changed
    type: list
    sample: [[{"Column": "Value1"},{"Column": "Value2"}], [{"ID": 1}, {"ID": 2}]]
rowcount:
    description: Number of affected rows for each subquery.
    returned: changed
    type: list
    sample: [5, 1]
execution_time_ms:
    description:
    - A list containing execution time per query in milliseconds.
    - The measurements are done right before and after passing
      the query to the driver for execution.
    returned: success
    type: list
    sample: [7104, 85]
    version_added: '3.12.0'
'''

import json
import time
import warnings

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.ansible.mysql.plugins.module_utils.mysql import (
    mysql_connect,
    mysql_common_argument_spec,
    mysql_driver,
    mysql_driver_fail_msg,
    set_session_vars,
)
from ansible.module_utils.common.text.converters import to_native

DML_QUERY_KEYWORDS = ('INSERT', 'UPDATE', 'DELETE', 'REPLACE')
# TRUNCATE is not DDL query but it also returns 0 rows affected:
DDL_QUERY_KEYWORDS = ('CREATE', 'DROP', 'ALTER', 'RENAME', 'TRUNCATE')


# ===========================================
# Module execution.
#

def get_time():
    try:
        time_taken = time.perf_counter()
    except AttributeError:
        # For Python 2 compatibility, fallback to time.time()
        time_taken = time.time()
    return time_taken


def execute_and_return_time(cursor, query, args):
    # Measure query execution time in milliseconds
    start_time = get_time()

    cursor.execute(query, args)

    # Calculate the execution time rounding it to 4 decimal places
    exec_time_ms = round((get_time() - start_time) * 1000, 4)
    return cursor, exec_time_ms


def main():
    argument_spec = mysql_common_argument_spec()
    argument_spec.update(
        query=dict(type='raw', required=True),
        login_db=dict(type='str'),
        positional_args=dict(type='list', elements='raw'),
        named_args=dict(type='dict'),
        single_transaction=dict(type='bool', default=False),
        session_vars=dict(type='dict'),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=(
            ('positional_args', 'named_args'),
        ),
    )

    db = module.params['login_db']
    connect_timeout = module.params['connect_timeout']
    login_user = module.params['login_user']
    login_password = module.params['login_password']
    ssl_cert = module.params['client_cert']
    ssl_key = module.params['client_key']
    ssl_ca = module.params['ca_cert']
    check_hostname = module.params['check_hostname']
    config_file = module.params['config_file']
    query = module.params["query"]
    session_vars = module.params["session_vars"]

    if not isinstance(query, (str, list)):
        module.fail_json(msg="the query option value must be a string or list, passed %s" % type(query))

    if isinstance(query, str):
        query = [query]

    for elem in query:
        if not isinstance(elem, str):
            module.fail_json(msg="the elements in query list must be strings, passed '%s' %s" % (elem, type(elem)))

    if module.params["single_transaction"]:
        autocommit = False
    else:
        autocommit = True
    # Prepare args:
    if module.params.get("positional_args"):
        arguments = module.params["positional_args"]
    elif module.params.get("named_args"):
        arguments = module.params["named_args"]
    else:
        arguments = None

    if mysql_driver is None:
        module.fail_json(msg=mysql_driver_fail_msg)

    # Connect to DB:
    try:
        cursor, db_connection = mysql_connect(module, login_user, login_password,
                                              config_file, ssl_cert, ssl_key, ssl_ca, db,
                                              check_hostname=check_hostname,
                                              connect_timeout=connect_timeout,
                                              cursor_class='DictCursor', autocommit=autocommit)
    except Exception as e:
        module.fail_json(msg="unable to connect to database, check login_user and "
                             "login_password are correct or %s has the credentials. "
                             "Exception message: %s" % (config_file, to_native(e)))

    # Set defaults:
    changed = False

    max_keyword_len = len(max(DML_QUERY_KEYWORDS + DDL_QUERY_KEYWORDS, key=len))

    if session_vars:
        set_session_vars(module, cursor, session_vars)

    # Execute query:
    query_result = []
    executed_queries = []
    rowcount = []
    execution_time_ms = []

    already_exists = False
    for q in query:
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings(action='error',
                                        message='.*already exists*',
                                        category=mysql_driver.Warning)

                try:
                    cursor, exec_time_ms = execute_and_return_time(cursor, q, arguments)
                    execution_time_ms.append(exec_time_ms)
                except mysql_driver.Warning:
                    # When something is run with IF NOT EXISTS
                    # and there's "already exists" MySQL warning,
                    # set the flag as True.
                    # PyMySQL < 0.10.0 throws the warning, mysqlclient
                    # and PyMySQL 0.10.0+ does NOT.
                    already_exists = True

        except Exception as e:
            if not autocommit:
                db_connection.rollback()

            cursor.close()
            module.fail_json(msg="Cannot execute SQL '%s' args [%s]: %s" % (q, arguments, to_native(e)))

        try:
            if not already_exists:
                rows = [dict(row) for row in cursor.fetchall()]
                # Round-trip through JSON to coerce non-serializable types
                # (e.g. decimal.Decimal, datetime) to their string representations,
                # preventing Ansible's exit_json from failing with
                # "Value of unknown type: <class 'decimal.Decimal'>".
                query_result.append(json.loads(json.dumps(rows, default=str)))

        except Exception as e:
            if not autocommit:
                db_connection.rollback()

            module.fail_json(msg="Cannot fetch rows from cursor: %s" % to_native(e))

        # Check DML or DDL keywords in query and set changed accordingly:
        q = q.lstrip()[0:max_keyword_len].upper()
        for keyword in DML_QUERY_KEYWORDS:
            if keyword in q and cursor.rowcount > 0:
                changed = True

        for keyword in DDL_QUERY_KEYWORDS:
            if keyword in q:
                if already_exists:
                    # Indicates the entity already exists
                    changed = False
                    already_exists = False  # Reset flag
                else:
                    changed = True
        try:
            executed_queries.append(cursor._last_executed)
        except AttributeError:
            # MySQLdb removed cursor._last_executed as a duplicate of cursor._executed
            executed_queries.append(cursor._executed)
        rowcount.append(cursor.rowcount)

    # When the module run with the single_transaction == True:
    if not autocommit:
        db_connection.commit()

    # Create dict with returned values:
    kw = {
        'changed': changed,
        'executed_queries': executed_queries,
        'query_result': query_result,
        'rowcount': rowcount,
        'execution_time_ms': execution_time_ms,
    }

    # Exit:
    module.exit_json(**kw)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2021, Ansible Project
# (c) 2019, XLAB d.o.o <www.xlab.si>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: s3_bucket_notification
version_added: 1.0.0
short_description: Creates, updates or deletes S3 Bucket notifications targeting Lambda functions, SNS or SQS.
description:
  - This module supports the creation, updates and deletions of S3 bucket notification profiles targeting
    either Lambda functions, SNS topics or SQS queues.
  - The target for the notifications must already exist. For lambdas use module M(community.aws.lambda)
    to manage the lambda function itself, M(community.aws.lambda_alias)
    to manage function aliases and M(community.aws.lambda_policy) to modify lambda permissions.
    For SNS or SQS then use M(community.aws.sns_topic) or M(community.aws.sqs_queue).
notes:
  - If using Lambda function as the target then a Lambda policy is also needed, use
    M(community.aws.lambda_policy) to do so to allow C(lambda:InvokeFunction) for the notification.
author:
  - XLAB d.o.o. (@xlab-si)
  - Aljaz Kosir (@aljazkosir)
  - Miha Plesko (@miha-plesko)
  - Mark Woolley (@marknet15)
options:
  event_name:
    description:
      - Unique name for event notification on bucket.
    required: true
    type: str
  bucket_name:
    description:
      - S3 bucket name.
    required: true
    type: str
  state:
    description:
      - Describes the desired state.
    default: "present"
    choices: ["present", "absent"]
    type: str
  queue_arn:
    description:
      - The ARN of the SQS queue.
      - Mutually exclusive with I(topic_arn) and I(lambda_function_arn).
    type: str
    version_added: 3.2.0
  topic_arn:
    description:
      - The ARN of the SNS topic.
      - Mutually exclusive with I(queue_arn) and I(lambda_function_arn).
    type: str
    version_added: 3.2.0
  lambda_function_arn:
    description:
      - The ARN of the lambda function.
      - Mutually exclusive with I(queue_arn) and I(topic_arn).
    aliases: ['function_arn']
    type: str
  lambda_alias:
    description:
      - Name of the Lambda function alias.
      - Mutually exclusive with I(lambda_version).
    type: str
  lambda_version:
    description:
      - Version of the Lambda function.
      - Mutually exclusive with I(lambda_alias).
    type: int
    default: 0
  events:
    description:
      - Events that will be triggering a notification. You can select multiple events to send
        to the same destination, you can set up different events to send to different destinations,
        and you can set up a prefix or suffix for an event. However, for each bucket,
        individual events cannot have multiple configurations with overlapping prefixes or
        suffixes that could match the same object key.
      - Required when I(state=present).
    choices: ['s3:ObjectCreated:*', 's3:ObjectCreated:Put', 's3:ObjectCreated:Post',
              's3:ObjectCreated:Copy', 's3:ObjectCreated:CompleteMultipartUpload',
              's3:ObjectRemoved:*', 's3:ObjectRemoved:Delete',
              's3:ObjectRemoved:DeleteMarkerCreated', 's3:ObjectRestore:Post',
              's3:ObjectRestore:Completed', 's3:ReducedRedundancyLostObject']
    type: list
    elements: str
    default: []
  prefix:
    description:
      - Optional prefix to limit the notifications to objects with keys that start with matching
        characters.
    type: str
    default: ''
  suffix:
    description:
      - Optional suffix to limit the notifications to objects with keys that end with matching
        characters.
    type: str
    default: ''
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
---
# Examples adding notification target configs to a S3 bucket
- name: Setup bucket event notification to a Lambda function
  community.aws.s3_bucket_notification:
    state: present
    event_name: on_file_add_or_remove
    bucket_name: test-bucket
    lambda_function_arn: arn:aws:lambda:us-east-2:123456789012:function:test-lambda
    events: ["s3:ObjectCreated:*", "s3:ObjectRemoved:*"]
    prefix: images/
    suffix: .jpg

- name: Setup bucket event notification to SQS
  community.aws.s3_bucket_notification:
    state: present
    event_name: on_file_add_or_remove
    bucket_name: test-bucket
    queue_arn: arn:aws:sqs:us-east-2:123456789012:test-queue
    events: ["s3:ObjectCreated:*", "s3:ObjectRemoved:*"]
    prefix: images/
    suffix: .jpg

# Example removing an event notification
- name: Remove event notification
  community.aws.s3_bucket_notification:
    state: absent
    event_name: on_file_add_or_remove
    bucket_name: test-bucket
"""

RETURN = r"""
notification_configuration:
  description: dictionary of currently applied notifications
  returned: success
  type: complex
  contains:
    lambda_function_configurations:
      description:
      - List of current Lambda function notification configurations applied to the bucket.
      type: list
    queue_configurations:
      description:
      - List of current SQS notification configurations applied to the bucket.
      type: list
    topic_configurations:
      description:
      - List of current SNS notification configurations applied to the bucket.
      type: list
"""

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    pass  # will be protected by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


class AmazonBucket:
    def __init__(self, module, client):
        self.module = module
        self.client = client
        self.bucket_name = module.params["bucket_name"]
        self.check_mode = module.check_mode
        self._full_config_cache = None

    def full_config(self):
        if self._full_config_cache is None:
            self._full_config_cache = dict(
                QueueConfigurations=[], TopicConfigurations=[], LambdaFunctionConfigurations=[]
            )

            try:
                config_lookup = self.client.get_bucket_notification_configuration(Bucket=self.bucket_name)
            except (ClientError, BotoCoreError) as e:
                self.module.fail_json(msg=f"{e}")

            # Handle different event targets
            if config_lookup.get("QueueConfigurations"):
                for queue_config in config_lookup.get("QueueConfigurations"):
                    self._full_config_cache["QueueConfigurations"].append(Config.from_api(queue_config))

            if config_lookup.get("TopicConfigurations"):
                for topic_config in config_lookup.get("TopicConfigurations"):
                    self._full_config_cache["TopicConfigurations"].append(Config.from_api(topic_config))

            if config_lookup.get("LambdaFunctionConfigurations"):
                for function_config in config_lookup.get("LambdaFunctionConfigurations"):
                    self._full_config_cache["LambdaFunctionConfigurations"].append(Config.from_api(function_config))

        return self._full_config_cache

    def current_config(self, config_name):
        # Iterate through configs and get current event config
        for target_configs in self.full_config():
            for config in self.full_config()[target_configs]:
                if config.raw["Id"] == config_name:
                    return config

    def apply_config(self, desired):
        configs = dict(QueueConfigurations=[], TopicConfigurations=[], LambdaFunctionConfigurations=[])

        # Iterate through existing configs then add the desired config
        for target_configs in self.full_config():
            for config in self.full_config()[target_configs]:
                if config.name != desired.raw["Id"]:
                    configs[target_configs].append(config.raw)

        if self.module.params.get("queue_arn"):
            configs["QueueConfigurations"].append(desired.raw)
        if self.module.params.get("topic_arn"):
            configs["TopicConfigurations"].append(desired.raw)
        if self.module.params.get("lambda_function_arn"):
            configs["LambdaFunctionConfigurations"].append(desired.raw)

        self._upload_bucket_config(configs)
        return configs

    def delete_config(self, desired):
        configs = dict(QueueConfigurations=[], TopicConfigurations=[], LambdaFunctionConfigurations=[])

        # Iterate through existing configs omitting specified config
        for target_configs in self.full_config():
            for config in self.full_config()[target_configs]:
                if config.name != desired.raw["Id"]:
                    configs[target_configs].append(config.raw)

        self._upload_bucket_config(configs)
        return configs

    def _upload_bucket_config(self, configs):
        api_params = dict(Bucket=self.bucket_name, NotificationConfiguration=dict())

        # Iterate through available configs
        for target_configs in configs:
            if len(configs[target_configs]) > 0:
                api_params["NotificationConfiguration"][target_configs] = configs[target_configs]

        if not self.check_mode:
            try:
                self.client.put_bucket_notification_configuration(**api_params)
            except (ClientError, BotoCoreError) as e:
                self.module.fail_json(msg=f"{e}")


class Config:
    def __init__(self, content):
        self._content = content
        self.name = content.get("Id")

    @property
    def raw(self):
        return self._content

    def __eq__(self, other):
        if other:
            return self.raw == other.raw
        return False

    @classmethod
    def from_params(cls, **params):
        """Generate bucket notification params for target"""

        bucket_event_params = dict(
            Id=params["event_name"],
            Events=sorted(params["events"]),
            Filter=dict(
                Key=dict(
                    FilterRules=[
                        dict(Name="Prefix", Value=params["prefix"]),
                        dict(Name="Suffix", Value=params["suffix"]),
                    ]
                )
            ),
        )

        # Handle different event targets
        if params.get("queue_arn"):
            bucket_event_params["QueueArn"] = params["queue_arn"]
        if params.get("topic_arn"):
            bucket_event_params["TopicArn"] = params["topic_arn"]
        if params.get("lambda_function_arn"):
            function_arn = params["lambda_function_arn"]

            qualifier = None
            if params["lambda_version"] > 0:
                qualifier = str(params["lambda_version"])
            elif params["lambda_alias"]:
                qualifier = str(params["lambda_alias"])
            if qualifier:
                params["lambda_function_arn"] = f"{function_arn}:{qualifier}"

            bucket_event_params["LambdaFunctionArn"] = params["lambda_function_arn"]

        return cls(bucket_event_params)

    @classmethod
    def from_api(cls, config):
        return cls(config)


def setup_module_object():
    event_types = [
        "s3:ObjectCreated:*",
        "s3:ObjectCreated:Put",
        "s3:ObjectCreated:Post",
        "s3:ObjectCreated:Copy",
        "s3:ObjectCreated:CompleteMultipartUpload",
        "s3:ObjectRemoved:*",
        "s3:ObjectRemoved:Delete",
        "s3:ObjectRemoved:DeleteMarkerCreated",
        "s3:ObjectRestore:Post",
        "s3:ObjectRestore:Completed",
        "s3:ReducedRedundancyLostObject",
    ]

    argument_spec = dict(
        state=dict(default="present", choices=["present", "absent"]),
        event_name=dict(required=True),
        lambda_function_arn=dict(aliases=["function_arn"]),
        queue_arn=dict(type="str"),
        topic_arn=dict(type="str"),
        bucket_name=dict(required=True),
        events=dict(type="list", default=[], choices=event_types, elements="str"),
        prefix=dict(default=""),
        suffix=dict(default=""),
        lambda_alias=dict(),
        lambda_version=dict(type="int", default=0),
    )

    mutually_exclusive = [
        ["queue_arn", "topic_arn", "lambda_function_arn"],
        ["lambda_alias", "lambda_version"],
    ]

    return AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=mutually_exclusive,
        required_if=[["state", "present", ["events"]]],
    )


def main():
    module = setup_module_object()

    client = module.client("s3")
    bucket = AmazonBucket(module, client)
    current = bucket.current_config(module.params["event_name"])
    desired = Config.from_params(**module.params)

    notification_configs = dict(QueueConfigurations=[], TopicConfigurations=[], LambdaFunctionConfigurations=[])

    for target_configs in bucket.full_config():
        for cfg in bucket.full_config()[target_configs]:
            notification_configs[target_configs].append(camel_dict_to_snake_dict(cfg.raw))

    state = module.params["state"]
    updated_configuration = dict()
    changed = False

    if state == "present":
        if current != desired:
            updated_configuration = bucket.apply_config(desired)
            changed = True
    elif state == "absent":
        if current:
            updated_configuration = bucket.delete_config(desired)
            changed = True

    for target_configs in updated_configuration:
        notification_configs[target_configs] = []
        for cfg in updated_configuration.get(target_configs, list()):
            notification_configs[target_configs].append(camel_dict_to_snake_dict(cfg))

    module.exit_json(changed=changed, notification_configuration=camel_dict_to_snake_dict(notification_configs))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Logging_global parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _tmplt_hosts(data):
    cmd = "logging server {host}"
    data["client_identity"] = data.get("secure", {}).get("trustpoint", {}).get("client_identity")

    if "severity" in data:
        cmd += " {severity}"
    if "port" in data:
        cmd += " port {port}"
    if data["client_identity"]:
        cmd += " secure trustpoint client-identity {client_identity}"
    if "facility" in data:
        cmd += " facility {facility}"
    if "use_vrf" in data:
        cmd += " use-vrf {use_vrf}"

    cmd = cmd.format(**data)

    return cmd


class Logging_globalTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Logging_globalTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "console",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\sconsole
                (\s(?P<severity>\d))?
                $""", re.VERBOSE,
            ),
            "setval": "{{ 'no ' if console.state|d('') == 'disabled' else '' }}"
                      "logging console"
                      "{{ (' ' + console.severity|string) if console.severity is defined else '' }}",
            "result": {
                "console": {
                    "state": "{{ 'disabled' if negated is defined else None }}",
                    "severity": "{{ severity }}",
                },
            },
        },
        {
            "name": "event.link_status.enable",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\sevent\slink-status\senable
                $""", re.VERBOSE,
            ),
            "setval": "logging event link-status enable",
            "result": {
                "event": {
                    "link_status": {
                        "enable": "{{ False if negated is defined else True }}",
                    },
                },
            },
        },
        {
            "name": "event.link_status.default",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\sevent\slink-status\sdefault
                $""", re.VERBOSE,
            ),
            "setval": "logging event link-status default",
            "result": {
                "event": {
                    "link_status": {
                        "default": "{{ False if negated is defined else True }}",
                    },
                },
            },
        },
        {
            "name": "event.trunk_status.enable",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\sevent\strunk-status\senable
                $""", re.VERBOSE,
            ),
            "setval": "logging event trunk-status enable",
            "result": {
                "event": {
                    "trunk_status": {
                        "enable": "{{ False if negated is defined else True }}",
                    },
                },
            },
        },
        {
            "name": "event.trunk_status.default",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\sevent\strunk-status\sdefault
                $""", re.VERBOSE,
            ),
            "setval": "logging event trunk-status default",
            "result": {
                "event": {
                    "trunk_status": {
                        "default": "{{ False if negated is defined else True }}",
                    },
                },
            },
        },
        {
            "name": "history.severity",
            "getval": re.compile(
                r"""
                ^logging\shistory
                \s(?P<severity>\d)
                $""", re.VERBOSE,
            ),
            "setval": "logging history {{ history.severity }}",
            "result": {
                "history": {
                    "severity": "{{ severity }}",
                },
            },
        },
        {
            "name": "history.size",
            "getval": re.compile(
                r"""
                ^logging\shistory\ssize
                \s(?P<size>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "logging history size {{ history.size }}",
            "result": {
                "history": {
                    "size": "{{ size }}",
                },
            },
        },
        {
            "name": "ip.access_list.cache.entries",
            "getval": re.compile(
                r"""
                ^logging\sip\saccess-list\scache
                \sentries\s(?P<entries>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "logging ip access-list cache entries {{ ip.access_list.cache.entries }}",
            "result": {
                "ip": {
                    "access_list": {
                        "cache": {
                            "entries": "{{ entries }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ip.access_list.cache.interval",
            "getval": re.compile(
                r"""
                ^logging\sip\saccess-list\scache
                \sinterval\s(?P<interval>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "logging ip access-list cache interval {{ ip.access_list.cache.interval }}",
            "result": {
                "ip": {
                    "access_list": {
                        "cache": {
                            "interval": "{{ interval }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ip.access_list.cache.threshold",
            "getval": re.compile(
                r"""
                ^logging\sip\saccess-list\scache
                \sthreshold\s(?P<threshold>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "logging ip access-list cache threshold {{ ip.access_list.cache.threshold }}",
            "result": {
                "ip": {
                    "access_list": {
                        "cache": {
                            "threshold": "{{ threshold }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ip.access_list.detailed",
            "getval": re.compile(
                r"""
                ^logging\sip\saccess-list
                \s(?P<detailed>detailed)
                $""", re.VERBOSE,
            ),
            "setval": "logging ip access-list detailed",
            "result": {
                "ip": {
                    "access_list": {
                        "detailed": "{{ not not detailed }}",
                    },
                },
            },
        },
        {
            "name": "ip.access_list.include.sgt",
            "getval": re.compile(
                r"""
                ^logging\sip\saccess-list\sinclude
                \s(?P<sgt>sgt)
                $""", re.VERBOSE,
            ),
            "setval": "logging ip access-list include sgt",
            "result": {
                "ip": {
                    "access_list": {
                        "include": {
                            "sgt": "{{ not not sgt }}",
                        },
                    },
                },
            },
        },
        {
            # in some cases, the `logging level` command
            # has an extra space at the end
            "name": "facilities",
            "getval": re.compile(
                r"""
                ^logging\slevel
                \s(?P<facility>\S+)
                \s(?P<severity>\d+)
                \s*
                $""", re.VERBOSE,
            ),
            "setval": "logging level {{ facility }} {{ severity }}",
            "result": {
                "facilities": [
                    {
                        "facility": "{{ facility }}",
                        "severity": "{{ severity }}",
                    },
                ],
            },
        },
        {
            "name": "logfile",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\slogfile
                (\s(?P<name>\S+))?
                (\s(?P<severity>\d+))?
                (\ssize\s(?P<size>\d+))?
                (\spersistent\sthreshold\s(?P<persistent_threshold>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "{{ 'no ' if logfile.state|d('') == 'disabled' else '' }}"
                      "logging logfile"
                      "{{ ' ' + logfile.name if logfile.name|d('') else '' }}"
                      "{{ (' ' + logfile.severity|string) if logfile.severity is defined else '' }}"
                      "{{ (' size ' + logfile.size|string) if logfile.size is defined else '' }}"
                      "{{ (' persistent threshold ' + logfile.persistent_threshold|string) if logfile.persistent_threshold is defined else '' }}",
            "result": {
                "logfile": {
                    "state": "{{ 'disabled' if negated is defined else None }}",
                    "name": "{{ name }}",
                    "severity": "{{ severity }}",
                    "persistent_threshold": "{{ persistent_threshold }}",
                    "size": "{{ size }}",
                },
            },
        },
        {
            "name": "module",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\smodule
                (\s(?P<severity>\d))?
                $""", re.VERBOSE,
            ),
            "setval": "{{ 'no ' if module.state|d('') == 'disabled' else '' }}"
                      "logging module"
                      "{{ (' ' + module.severity|string) if module.severity is defined else '' }}",
            "result": {
                "module": {
                    "state": "{{ 'disabled' if negated is defined else None }}",
                    "severity": "{{ severity }}",
                },
            },
        },
        {
            "name": "monitor",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging\smonitor
                (\s(?P<severity>\d))?
                $""", re.VERBOSE,
            ),
            "setval": "{{ 'no ' if monitor.state|d('') == 'disabled' else '' }}"
                      "logging monitor"
                      "{{ (' ' + monitor.severity|string) if monitor.severity is defined else '' }}",
            "result": {
                "monitor": {
                    "state": "{{ 'disabled' if negated is defined else None }}",
                    "severity": "{{ severity }}",
                },
            },
        },
        {
            "name": "origin_id.hostname",
            "getval": re.compile(
                r"""
                ^logging\sorigin-id
                \s(?P<hostname>hostname)
                $""", re.VERBOSE,
            ),
            "setval": "logging origin-id hostname",
            "result": {
                "origin_id": {
                    "hostname": "{{ not not hostname }}",
                },
            },
        },
        {
            "name": "origin_id.ip",
            "getval": re.compile(
                r"""
                ^logging\sorigin-id
                \sip\s(?P<ip>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "logging origin-id ip {{ origin_id.ip }}",
            "result": {
                "origin_id": {
                    "ip": "{{ ip }}",
                },
            },
        },
        {
            "name": "origin_id.string",
            "getval": re.compile(
                r"""
                ^logging\sorigin-id
                \sstring\s(?P<string>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "logging origin-id string {{ origin_id.string }}",
            "result": {
                "origin_id": {
                    "string": "{{ string }}",
                },
            },
        },
        {
            "name": "rate_limit",
            "getval": re.compile(
                r"""
                ^(?P<negated>no\s)?
                logging
                \s(?P<rate_limit>rate-limit)
                $""", re.VERBOSE,
            ),
            "setval": "{{ 'no ' if rate_limit|d('') == 'disabled' else '' }}"
                      "logging rate-limit",
            "result": {
                "rate_limit": "{{ 'disabled' if negated is defined else None }}",
            },
        },
        {
            "name": "rfc_strict",
            "getval": re.compile(
                r"""
                logging\srfc-strict
                \s(?P<rfc_strict>5424)
                $""", re.VERBOSE,
            ),
            "setval": "logging rfc-strict 5424",
            "result": {
                "rfc_strict": "{{ not not rfc_strict }}",
            },
        },
        {
            "name": "hosts",
            "getval": re.compile(
                r"""
                ^logging\sserver
                \s(?P<host>\S+)
                (\s(?P<severity>\d))?
                (\sport\s(?P<port>\d+))?
                (\ssecure\strustpoint\sclient-identity\s(?P<client_identity>\S+))?
                (\suse-vrf\s(?P<use_vrf>\S+))?
                (\sfacility\s(?P<facility>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": _tmplt_hosts,
            "result": {
                "hosts": [
                    {
                        "host": "{{ host }}",
                        "severity": "{{ severity }}",
                        "secure": {
                            "trustpoint": {
                                "client_identity": "{{ client_identity }}",
                            },
                        },
                        "port": "{{ port }}",
                        "facility": "{{ facility }}",
                        "use_vrf": "{{ use_vrf }}",
                    },
                ],
            },
        },
        {
            "name": "source_interface",
            "getval": re.compile(
                r"""
                ^logging\ssource-interface
                \s(?P<source_interface>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "logging source-interface {{ source_interface }}",
            "result": {
                "source_interface": "{{ source_interface }}",
            },
        },
        {
            "name": "timestamp",
            "getval": re.compile(
                r"""
                ^logging\stimestamp
                \s(?P<timestamp>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "logging timestamp {{ timestamp }}",
            "result": {
                "timestamp": "{{ timestamp }}",
            },
        },
    ]
    # fmt: on



#!/usr/bin/python

# Copyright (c) 2013, Jimmy Tang <jcftang@gmail.com>
# Based on okpg (Patrick Pelletier <pp.pelletier@gmail.com>), pacman
# (Afterburn) and pkgin (Shaun Zinck) modules
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: macports
author: "Jimmy Tang (@jcftang)"
short_description: Package manager for MacPorts
description:
  - Manages MacPorts packages (ports).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    description:
      - A list of port names.
    aliases: ['port']
    type: list
    elements: str
  selfupdate:
    description:
      - Update Macports and the ports tree, either prior to installing ports or as a separate step.
      - Equivalent to running C(port selfupdate).
    aliases: ['update_cache', 'update_ports']
    default: false
    type: bool
  state:
    description:
      - Indicates the desired state of the port.
    choices: ['present', 'absent', 'active', 'inactive', 'installed', 'removed']
    default: present
    type: str
  upgrade:
    description:
      - Upgrade all outdated ports, either prior to installing ports or as a separate step.
      - Equivalent to running C(port upgrade outdated).
    default: false
    type: bool
  variant:
    description:
      - A port variant specification.
      - O(variant) is only supported with O(state=installed) and O(state=present).
    aliases: ['variants']
    type: str
"""

EXAMPLES = r"""
- name: Install the foo port
  community.general.macports:
    name: foo

- name: Install the universal, x11 variant of the foo port
  community.general.macports:
    name: foo
    variant: +universal+x11

- name: Install a list of ports
  community.general.macports:
    name: "{{ ports }}"
  vars:
    ports:
      - foo
      - foo-tools

- name: Update Macports and the ports tree, then upgrade all outdated ports
  community.general.macports:
    selfupdate: true
    upgrade: true

- name: Update Macports and the ports tree, then install the foo port
  community.general.macports:
    name: foo
    selfupdate: true

- name: Remove the foo port
  community.general.macports:
    name: foo
    state: absent

- name: Activate the foo port
  community.general.macports:
    name: foo
    state: active

- name: Deactivate the foo port
  community.general.macports:
    name: foo
    state: inactive
"""

import re

from ansible.module_utils.basic import AnsibleModule


def selfupdate(module, port_path):
    """Update Macports and the ports tree."""

    rc, out, err = module.run_command([port_path, "-v", "selfupdate"])

    if rc == 0:
        updated = any(
            re.search(r"Total number of ports parsed:\s+[^0]", s.strip())
            or re.search(r"Installing new Macports release", s.strip())
            for s in out.split("\n")
            if s
        )
        if updated:
            changed = True
            msg = "Macports updated successfully"
        else:
            changed = False
            msg = "Macports already up-to-date"

        return (changed, msg, out, err)
    else:
        module.fail_json(msg="Failed to update Macports", stdout=out, stderr=err)


def upgrade(module, port_path):
    """Upgrade outdated ports."""

    rc, out, err = module.run_command([port_path, "upgrade", "outdated"])

    # rc is 1 when nothing to upgrade so check stdout first.
    if out.strip() == "Nothing to upgrade.":
        changed = False
        msg = "Ports already upgraded"
        return (changed, msg, out, err)
    elif rc == 0:
        changed = True
        msg = "Outdated ports upgraded successfully"
        return (changed, msg, out, err)
    else:
        module.fail_json(msg="Failed to upgrade outdated ports", stdout=out, stderr=err)


def query_port(module, port_path, name, state="present"):
    """Returns whether a port is installed or not."""

    if state == "present":
        rc, out, err = module.run_command([port_path, "-q", "installed", name])

        return rc == 0 and out.strip().startswith(f"{name} ")

    elif state == "active":
        rc, out, err = module.run_command([port_path, "-q", "installed", name])

        return rc == 0 and "(active)" in out


def remove_ports(module, port_path, ports, stdout, stderr):
    """Uninstalls one or more ports if installed."""

    remove_c = 0
    # Using a for loop in case of error, we can report the port that failed
    for port in ports:
        # Query the port first, to see if we even need to remove
        if not query_port(module, port_path, port):
            continue

        rc, out, err = module.run_command([port_path, "uninstall", port])
        stdout += out
        stderr += err
        if query_port(module, port_path, port):
            module.fail_json(msg=f"Failed to remove {port}: {err}", stdout=stdout, stderr=stderr)

        remove_c += 1

    if remove_c > 0:
        module.exit_json(changed=True, msg=f"Removed {remove_c} port(s)", stdout=stdout, stderr=stderr)

    module.exit_json(changed=False, msg="Port(s) already absent", stdout=stdout, stderr=stderr)


def install_ports(module, port_path, ports, variant, stdout, stderr):
    """Installs one or more ports if not already installed."""

    install_c = 0

    for port in ports:
        if query_port(module, port_path, port):
            continue

        rc, out, err = module.run_command([port_path, "install", port, variant])
        stdout += out
        stderr += err
        if not query_port(module, port_path, port):
            module.fail_json(msg=f"Failed to install {port}: {err}", stdout=stdout, stderr=stderr)

        install_c += 1

    if install_c > 0:
        module.exit_json(changed=True, msg=f"Installed {install_c} port(s)", stdout=stdout, stderr=stderr)

    module.exit_json(changed=False, msg="Port(s) already present", stdout=stdout, stderr=stderr)


def activate_ports(module, port_path, ports, stdout, stderr):
    """Activate a port if it is inactive."""

    activate_c = 0

    for port in ports:
        if not query_port(module, port_path, port):
            module.fail_json(msg=f"Failed to activate {port}, port(s) not present", stdout=stdout, stderr=stderr)

        if query_port(module, port_path, port, state="active"):
            continue

        rc, out, err = module.run_command([port_path, "activate", port])
        stdout += out
        stderr += err

        if not query_port(module, port_path, port, state="active"):
            module.fail_json(msg=f"Failed to activate {port}: {err}", stdout=stdout, stderr=stderr)

        activate_c += 1

    if activate_c > 0:
        module.exit_json(changed=True, msg=f"Activated {activate_c} port(s)", stdout=stdout, stderr=stderr)

    module.exit_json(changed=False, msg="Port(s) already active", stdout=stdout, stderr=stderr)


def deactivate_ports(module, port_path, ports, stdout, stderr):
    """Deactivate a port if it is active."""

    deactivated_c = 0

    for port in ports:
        if not query_port(module, port_path, port):
            module.fail_json(msg=f"Failed to deactivate {port}, port(s) not present", stdout=stdout, stderr=stderr)

        if not query_port(module, port_path, port, state="active"):
            continue

        rc, out, err = module.run_command([port_path, "deactivate", port])
        stdout += out
        stderr += err
        if query_port(module, port_path, port, state="active"):
            module.fail_json(msg=f"Failed to deactivate {port}: {err}", stdout=stdout, stderr=stderr)

        deactivated_c += 1

    if deactivated_c > 0:
        module.exit_json(changed=True, msg=f"Deactivated {deactivated_c} port(s)", stdout=stdout, stderr=stderr)

    module.exit_json(changed=False, msg="Port(s) already inactive", stdout=stdout, stderr=stderr)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="list", elements="str", aliases=["port"]),
            selfupdate=dict(aliases=["update_cache", "update_ports"], default=False, type="bool"),
            state=dict(default="present", choices=["present", "installed", "absent", "removed", "active", "inactive"]),
            upgrade=dict(default=False, type="bool"),
            variant=dict(aliases=["variants"], type="str"),
        )
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    stdout = ""
    stderr = ""

    port_path = module.get_bin_path("port", True, ["/opt/local/bin"])

    p = module.params

    if p["selfupdate"]:
        (changed, msg, out, err) = selfupdate(module, port_path)
        stdout += out
        stderr += err
        if not (p["name"] or p["upgrade"]):
            module.exit_json(changed=changed, msg=msg, stdout=stdout, stderr=stderr)

    if p["upgrade"]:
        (changed, msg, out, err) = upgrade(module, port_path)
        stdout += out
        stderr += err
        if not p["name"]:
            module.exit_json(changed=changed, msg=msg, stdout=stdout, stderr=stderr)

    pkgs = p["name"]

    variant = p["variant"]

    if p["state"] in ["present", "installed"]:
        install_ports(module, port_path, pkgs, variant, stdout, stderr)

    elif p["state"] in ["absent", "removed"]:
        remove_ports(module, port_path, pkgs, stdout, stderr)

    elif p["state"] == "active":
        activate_ports(module, port_path, pkgs, stdout, stderr)

    elif p["state"] == "inactive":
        deactivate_ports(module, port_path, pkgs, stdout, stderr)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_validate
short_description: Validate the schema before deploying it to site
description:
- This module is used to verify if a schema can be deployed to site without any error.
- This module can only be used on versions of MSO that are 3.3 or greater.
- Starting with MSO 3.3, the schema modules in this collection will skip some validation checks to allow part of the schema to be updated more easily.
- This module will check those validation after all changes have been made.
author:
- Anvitha Jain (@anvitha-jain)
version_added: "1.3.0"
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  state:
    description:
    - Use C(query) to validate deploying the schema.
    type: str
    default: query
    choices: [ query ]
seealso:
- module: cisco.mso.mso_schema
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Get Validation status
  mso_schema_validate:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    state: query
  register: query_validate
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        state=dict(type="str", default="query", choices=["query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    schema = module.params.get("schema")

    mso = MSOModule(module)

    mso.existing = mso.validate_schema(schema_id=mso.lookup_schema(schema))

    mso.exit_json()


if __name__ == "__main__":
    main()



from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible_collections.cisco.mso.plugins.module_utils.schema import MSOSchema


class MSOSchemas:
    def __init__(self, mso_module):
        self.mso = mso_module
        self.schemas_by_id = {}
        self.schemas_by_name = {}

    def get_template_from_schema(self, schema_name, schema_id, template_name, template_id, refresh=False):
        if not refresh:
            existing_schema = None

            if schema_id in self.schemas_by_id:
                existing_schema = self.schemas_by_id[schema_id]
            elif schema_name in self.schemas_by_name:
                existing_schema = self.schemas_by_name[schema_name]

            if existing_schema:
                if template_id:
                    existing_schema.set_template_from_id(template_id)
                elif template_name:
                    existing_schema.set_template(template_name)
                return existing_schema

        new_schema = MSOSchema(self.mso, schema_name, template_name, None, schema_id, template_id)
        self.schemas_by_id[new_schema.id] = new_schema
        self.schemas_by_name[new_schema.schema_name] = new_schema
        return new_schema

    def get_epg_uuid(self, epg_obj, epg_uuid):
        """
        Returns the EPG UUID in a given schema, template and ANP.

        :param epg_obj: A dictionary containing the epg reference object. -> Dict
        :param epg_uuid: The EPG UUID to return if specified. -> Str
        :return: The EPG UUID if found, otherwise it fails the module. -> Str
        """
        if epg_uuid:
            return epg_uuid
        schema = self.get_template_from_schema(
            epg_obj.get("schema"),
            epg_obj.get("schema_id"),
            epg_obj.get("template"),
            epg_obj.get("template_id"),
        )
        schema.set_template_anp(epg_obj.get("anp"), epg_obj.get("anp_uuid"), fail_module=True)
        schema.set_template_anp_epg(epg_obj.get("name"), fail_module=True)
        return schema.schema_objects.get("template_anp_epg").details.get("uuid")



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
    author: Ganesh Nalawade (@ganeshrn)
    name: jsonschema
    short_description: Define configurable options for jsonschema validate plugin
    description:
    - This sub plugin documentation provides the configurable options that can be passed
      to the validate plugins when C(ansible.utils.jsonschema) is used as a value for
      engine option.
    version_added: 1.0.0
    options:
      draft:
        description:
        - This option provides the jsonschema specification that should be used
          for the validating the data. The I(criteria) option in the validate
          plugin should follow the specification as mentioned by this option.
          If this option is not specified, jsonschema will use the best validator
          for the I($schema) field in the criteria. Specifications 2019-09 and
          2020-12 are only available from jsonschema version 4.0 onwards.
        choices:
        - draft3
        - draft4
        - draft6
        - draft7
        - 2019-09
        - 2020-12
        env:
        - name: ANSIBLE_VALIDATE_JSONSCHEMA_DRAFT
        vars:
        - name: ansible_validate_jsonschema_draft
      check_format:
        description: If enabled, validate the I(format) specification in the criteria.
        type: bool
        default: true
        env:
        - name: ANSIBLE_VALIDATE_JSONSCHEMA_CHECK_FORMAT
        vars:
        - name: ansible_validate_jsonschema_check_format
    notes:
    - The value of I(data) option should be either a valid B(JSON) object or a B(JSON) string.
    - The value of I(criteria) should be B(list) of B(dict) or B(list) of B(strings) and each
      B(string) within the B(list) entry should be a valid B(dict) when read in python.
"""

import json

from ansible.errors import AnsibleError
from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_text
from ansible.utils.display import Display

from ansible_collections.ansible.utils.plugins.module_utils.common.utils import to_list
from ansible_collections.ansible.utils.plugins.plugin_utils.base.validate import ValidateBase


display = Display()

# PY2 compatibility for JSONDecodeError
try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

try:
    import jsonschema
    import jsonschema.validators

    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False


def to_path(fpath):
    return ".".join(str(index) for index in fpath)


def json_path(absolute_path):
    path = "$"
    for elem in absolute_path:
        if isinstance(elem, int):
            path += "[" + str(elem) + "]"
        else:
            path += "." + elem
    return path


class Validate(ValidateBase):
    # All available schema versions with the format_check and validator class names.
    _JSONSCHEMA_DRAFTS = {
        "draft3": {
            "validator_name": "Draft3Validator",
            "format_checker_name": "draft3_format_checker",
        },
        "draft4": {
            "validator_name": "Draft4Validator",
            "format_checker_name": "draft4_format_checker",
        },
        "draft6": {
            "validator_name": "Draft6Validator",
            "format_checker_name": "draft6_format_checker",
        },
        "draft7": {
            "validator_name": "Draft7Validator",
            "format_checker_name": "draft7_format_checker",
        },
        "2019-09": {
            "validator_name": "Draft201909Validator",
            "format_checker_name": "draft201909_format_checker",
        },
        "2020-12": {
            "validator_name": "Draft202012Validator",
            "format_checker_name": "draft202012_format_checker",
        },
    }

    @staticmethod
    def _check_reqs():
        """Check the prerequisites are installed for jsonschema

        :return None: In case all prerequisites are satisfied
        """
        if not HAS_JSONSCHEMA:
            raise AnsibleError(missing_required_lib("jsonschema"))

    def _check_args(self):
        """Ensure specific args are set

        :return: None: In case all arguments passed are valid
        """
        try:
            if isinstance(self._data, str):
                self._data = json.loads(self._data)
            else:
                self._data = json.loads(json.dumps(self._data))

        except (TypeError, JSONDecodeError) as exe:
            msg = (
                "'data' option value is invalid, value should a valid JSON."
                " Failed to read with error '{err}'".format(
                    err=to_text(exe, errors="surrogate_then_replace"),
                )
            )
            raise AnsibleError(msg)

        try:
            criteria = []
            for item in to_list(self._criteria):
                if isinstance(self._criteria, str):
                    criteria.append(json.loads(item))
                else:
                    criteria.append(json.loads(json.dumps(item)))

            self._criteria = criteria
        except (TypeError, JSONDecodeError) as exe:
            msg = (
                "'criteria' option value is invalid, value should a valid JSON."
                " Failed to read with error '{err}'".format(
                    err=to_text(exe, errors="surrogate_then_replace"),
                )
            )
            raise AnsibleError(msg)

    def _check_drafts(self):
        """For every possible draft check if our jsonschema version supports it and exchange the class names with
        the actual classes. If it is not supported the draft is removed from the list.
        """
        for draft in list(self._JSONSCHEMA_DRAFTS.keys()):
            draft_config = self._JSONSCHEMA_DRAFTS[draft]
            try:
                validator_class = getattr(jsonschema, draft_config["validator_name"])
            except AttributeError:
                display.vvv(
                    'jsonschema draft "{draft}" not supported in this version'.format(draft=draft),
                )
                del self._JSONSCHEMA_DRAFTS[draft]
                continue
            draft_config["validator"] = validator_class
            try:
                format_checker_class = validator_class.FORMAT_CHECKER
            except AttributeError:
                # Older jsonschema version
                format_checker_class = getattr(jsonschema, draft_config["format_checker_name"])
            draft_config["format_checker"] = format_checker_class

    def validate(self):
        """Std entry point for a validate execution

        :return: Errors or parsed text as structured data
        :rtype: dict

        :example:

        The parse function of a parser should return a dict:
        {"errors": [a list of errors]}
        or
        {"parsed": obj}
        """
        self._check_reqs()
        self._check_args()
        self._check_drafts()

        try:
            self._validate_jsonschema()
        except Exception as exc:
            return {"errors": to_text(exc, errors="surrogate_then_replace")}

        return self._result

    def _validate_jsonschema(self):
        error_messages = None

        draft = self._get_sub_plugin_options("draft")
        check_format = self._get_sub_plugin_options("check_format")
        error_messages = []

        for criteria in self._criteria:
            format_checker = None
            validator_class = None
            if draft is not None:
                try:
                    validator_class = self._JSONSCHEMA_DRAFTS[draft]["validator"]
                except KeyError:
                    display.warning(
                        'No validator available for "{draft}", falling back to autodetection. A newer version of jsonschema might support this draft.'.format(
                            draft=draft,
                        ),
                    )
            if validator_class is None:
                # Either no draft was specified or specified draft has no validator class
                # in installed jsonschema version. Do autodetection instead.
                validator_class = jsonschema.validators.validator_for(criteria)

            if check_format:
                try:
                    format_checker = validator_class.FORMAT_CHECKER
                except AttributeError:
                    # TODO: Remove when Python 3.6 support is dropped.
                    # On jsonschema<4.5, there is no connection between a validator and the correct format checker.
                    # So we iterate through our known list of validators and if one matches the current class
                    # we use the format_checker from that validator.
                    for draft, draft_config in self._JSONSCHEMA_DRAFTS.items():
                        if validator_class == draft_config["validator"]:
                            display.vvv(
                                "Using format_checker for {draft} validator".format(draft=draft),
                            )
                            format_checker = draft_config["format_checker"]
                            break
                    else:
                        display.warning("jsonschema format checks not available")

            validator = validator_class(criteria, format_checker=format_checker)
            validation_errors = sorted(validator.iter_errors(self._data), key=lambda e: e.path)

            if validation_errors:
                if "errors" not in self._result:
                    self._result["errors"] = []

                for validation_error in validation_errors:
                    if isinstance(validation_error, jsonschema.ValidationError):
                        error = {
                            "message": validation_error.message,
                            "data_path": to_path(validation_error.absolute_path),
                            "json_path": json_path(validation_error.absolute_path),
                            "schema_path": to_path(validation_error.relative_schema_path),
                            "relative_schema": validation_error.schema,
                            "expected": validation_error.validator_value,
                            "validator": validation_error.validator,
                            "found": validation_error.instance,
                        }
                        self._result["errors"].append(error)
                        error_message = "At '{schema_path}' {message}. ".format(
                            schema_path=error["schema_path"],
                            message=error["message"],
                        )
                        error_messages.append(error_message)
        if error_messages:
            if "msg" not in self._result:
                self._result["msg"] = "\n".join(error_messages)
            else:
                self._result["msg"] += "\n".join(error_messages)
